/* Enhanced Communication Transport Protocol Part-3 Implementation
 *
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * This header defines the ECTP-3 Internal Functions used to Protocol
 * Operations.
 *
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers	<saintpaul1978@mail.knu.ac.kr>
 *						<knucsid@gmail.com>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 *	Jae Sung Park				<knucsid@gmail.com>
 *
 */

#include <ectp/types.h>
#include <ectp/structs.h>
#include <ectp/protocol.h>


#ifndef	_ECTP_H
#define	_ECTP_H

#ifdef	__cplusplus
extern "C" {
#endif


/* ------------------------------------------------------------------------- *
 * EVENT INVOCATION FUNCTIONS
 * ------------------------------------------------------------------------- */

/* control.c --------------------------------------------------------------- */

extern int ectp_ops_start_events( struct ectp_socket *eso );
extern int ectp_ops_end_events( struct ectp_socket *eso );



/* ------------------------------------------------------------------------- *
 * STATE MACHINE CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

/* control.c --------------------------------------------------------------- */

/* Start Connection Creation handshake. */
extern int ectp_ops_connection_creation( struct ectp_socket *eso, int time );

/* Start Late Join handshake. */
extern int ectp_ops_start_latejoin( struct ectp_socket *eso, int time );

extern int ectp_ops_user_leave( struct ectp_socket *eso,
								ectp_addr_t *to, socklen_t tolen );

extern int ectp_ops_request_userleave( struct ectp_socket *eso,
								ectp_addr_t *from, socklen_t fromlen );

extern int ectp_ops_conn_terminate( struct ectp_socket *eso, int f );
 
extern int ectp_ops_request_tokenget( struct ectp_socket *eso );
extern int ectp_ops_request_tokenreturn( struct ectp_socket *eso );

extern int ectp_ops_ack( struct ectp_socket *eso );

/* By using POSIX thread, all of following operations can run in ESTABLISHED
 * state. Therefore, there is no change in state machine. */

extern int ectp_ops_accept_latejoin(
		struct ectp_socket *eso, ectp_addr_t *from );

extern int ectp_ops_deny_latejoin(
		struct ectp_socket *eso, ectp_addr_t *from );

extern int ectp_ops_accept_treejoin(
		struct ectp_socket *eso, ectp_addr_t *from );

extern int ectp_ops_deny_treejoin(
		struct ectp_socket *eso, ectp_addr_t *from );

/**
 * TO must specify a new valid token.
 */
extern int ectp_ops_accept_tokenget(
		struct ectp_socket *eso, ectp_addr_t *from );

extern int ectp_ops_deny_tokenget(
		struct ectp_socket *eso, ectp_addr_t *from );

extern int ectp_ops_accept_tokenreturn(
		struct ectp_socket *eso, ectp_token_t token, ectp_addr_t *from );



/* ------------------------------------------------------------------------- *
 * ERROR CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

/**
 * Generate checksum.
 */
static inline __u16 ectp_checksum( __u16 *buffer, int size )
{
	__u32 cksum = 0;

	while( size > 1 )
	{
		cksum += *buffer++;
		size -= sizeof(__u16);
	}

	// if size is odd. ex) buffer ==> char buf[3]
	if( size )
		cksum += *(__u8 * )buffer;

	cksum = ( cksum >> 16 ) + ( cksum & 0xffff );
	cksum += ( cksum >> 16 );
	return (__u16) (~cksum);
}



/* ------------------------------------------------------------------------- *
 * EVENT/TIMER CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

/* event.c ----------------------------------------------------------------- */

extern void ectp_event_dispatch(struct ectp_socket *eso, fd_set *fds);

/* timer.c ----------------------------------------------------------------- */

extern void ectp_timer_dispatch( struct ectp_socket *eso );



/* ------------------------------------------------------------------------- *
 * INPUT/OUTPUT CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

extern ectp_pkt_ct *ectp_create_ct( struct ectp_socket *eso, int f );
extern ectp_pkt_lr *ectp_create_lr( struct ectp_socket *eso, int f );

extern ectp_pkt_dt *ectp_create_dt( struct ectp_socket *eso,
									int payloadlen, int flags);
extern ectp_pkt_ack *ectp_create_ack( struct ectp_socket *eso );

extern ectp_pkt_jr *ectp_create_jr( struct ectp_socket *eso );
extern ectp_pkt_jc *ectp_create_jc( struct ectp_socket *eso,
									int f, int option );

extern ectp_pkt_tj *ectp_create_tj( struct ectp_socket *eso );
extern ectp_pkt_tc *ectp_create_tc( struct ectp_socket *eso,
									int f, int option );

extern ectp_pkt_tgr *ectp_create_tgr( struct ectp_socket *eso );
extern ectp_pkt_tgc *ectp_create_tgc( struct ectp_socket *eso, int f );

extern ectp_pkt_trr *ectp_create_trr( struct ectp_socket *eso, int token_id );
extern ectp_pkt_trc *ectp_create_trc( struct ectp_socket *eso, int f );

/**
 * @param length	indicates the length of optional fields or user data.
 * @param domain	indicates whether the address family is IPv4 or IPv6.
 * 					If domain is zero, it will be ignored.
 */
void *ectp_pkt_alloc(ectp_packet_t type, int length, int domain);

static inline ssize_t ectp_pkt_transmit( int sd, void *pkt, size_t pktlen,
										 ectp_addr_t *to, socklen_t tolen )
{
	struct ectp_basehdr *basehdr = (struct ectp_basehdr *) pkt;
	basehdr->checksum = ectp_checksum((__u16 *) pkt, pktlen);

	return sendto(sd, pkt, pktlen, 0, (SA *) to, tolen);
}

/**
 * ectp_skb_transmit() generate checksum for the given packet in the send
 * buffer, and transmit it to the destination endpoint.
 *
 * @param sd	internal udp socket descriptor to send
 */
static inline ssize_t ectp_skb_transmit( struct ectp_skbuff *skb )
{
	struct ectp_basehdr *basehdr = (struct ectp_basehdr *) skb->pkt;
	basehdr->checksum = ectp_checksum((__u16 *) skb->pkt, skb->pktlen);
	skb->has_sent = 1;

	return sendto(skb->sd, skb->pkt, skb->pktlen, 0,
				  (SA *) &skb->addr, skb->addrlen);
}

/** retransmit -- don't have to calculate checksum and ... */
static inline ssize_t ectp_skb_retransmit( struct ectp_skbuff *skb )
{
	skb->has_sent = 1;
	return sendto(skb->sd, skb->pkt, skb->pktlen, 0,
				  (SA *) &skb->addr, skb->addrlen);
}

extern ssize_t ectp_ops_recvmsg(struct ectp_socket *eso,
							void *msg, size_t len, int flags,
							ectp_addr_t *from, socklen_t *fromlen);

extern ssize_t ectp_ops_sendmsg(struct ectp_socket *eso,
							const void *msg, size_t len, int flags,
							ectp_addr_t *to, socklen_t tolen);

/* Flags for send/receive operations. */
#define	ECTP_UCAST		0x01	/* Transmit Unicast Data */
#define	ECTP_MCAST		0x02	/* Transmit Multicast Data */
#define	ECTP_DONTWAIT	0x10	/* Nonblocking I/O */


#ifdef	__cplusplus
}
#endif

#endif	/* _ECTP_H */
