
#include <ectp/types.h>

#ifndef	_ECTP_EVENT_H
#define _ECTP_EVENT_H

#ifdef	__cplusplus
extern "C" {
#endif


/* ------------------------------------------------------------------------- *
 * Definitions for ECTP-3 User Callback
 * ------------------------------------------------------------------------- */

/**
 * ectp_callback{} contaions two function pointers that are given from TO
 * application (usually a server).
 */
struct ectp_callback {
	
	/* Late Join Handler */
	void (*latejoin_handler) (int esd, SS *from, socklen_t fromlen);

	/* Tree Join Handler */
	/* void (*treejoin_handler) (SS *from, socklen_t *fromlen); */

	/* Token Get Handler */
	void (*tokenget_handler) (int esd, SS *from, socklen_t fromlen);
};

/* event.c ----------------------------------------------------------------- */

extern void ectp_init_callback( struct ectp_callback *callback );

extern void ectp_update_callback( struct ectp_callback *dst,
								  struct ectp_callback *src );



/* ------------------------------------------------------------------------- *
 * Definitions for ECTP-3 Internal Event Loop & Handlers (Thread Functons)
 * ------------------------------------------------------------------------- */


/**
 * ECTP-3 Internal Event Handlers (Posix Thread functions)
 *
 * NOTE:
 * 		DO NOT CALL EVENT HANDLERS DIRECTLY.
 */ 
extern void *ectp_main_loop( void *arg );
extern void *ectp_join_handler( void *arg );
extern void *ectp_token_handler( void *arg );
extern void *ectp_event_handler( void *arg );

#ifndef	_ECTP_DEBUG

#define	ECTP_MAINLOOP_TIMEOUT_SEC	0
#define	ECTP_MAINLOOP_TIMEOUT_USEC	100000	/* 100 [msec] */

#elif defined (_ECTP_DEBUG)

#define ECTP_MAINLOOP_TIMEOUT_SEC	10		/* 1 [sec] */
#define	ECTP_MAINLOOP_TIMEOUT_USEC	0

#endif


/**
 * Internal Event Handle Set.
 * - This structure is the collection of internal event & its handlers.
 *
 * NOTE:
 * 		We divide one event handler into three because of concurrency.
 */
struct ectp_handleset
{
	/* These fields represents whether our ECTP-3 event handlers are running.
	 * USAGE:
	 * 		while ( eso->hset.handler_is_running ) {
	 * 			// some event handler operations.
	 * 		}
	 *
	 */
	__u8		main_loop_is_running;
	__u8		event_handler_is_running;
	__u8		join_handler_is_running;
	__u8		token_handler_is_running;

	pthread_t	loop;	/* Thread ID */
	void		*(*main_loop)		(void *arg);
	pthread_t	event;
	void		*(*event_handler)	(void *arg);

	/* Following fields are not necessary to TU. */
	pthread_t	join;
	void		*(*join_handler)	(void *arg);

	pthread_t	token;
	void		*(*token_handler)	(void *arg);
};

/* event.c ----------------------------------------------------------------- */

extern void ectp_handleset_init( struct ectp_handleset *hset,
								 ectp_role_t role );


/* ------------------------------------------------------------------------- *
 * Definitions for ECTP-3 Internal Event Queue
 * ------------------------------------------------------------------------- */

/**
 * ECTP-3 Events Description.
 *
 * -	Request event:	ECTP_REQ_xxx
 *  	Confirm event:	ECTP_CON_xxx
 *  	Deny    event:	ECTP_DNY_xxx
 *  	Expire  time :	ECTP_EXP_xxx
 */ 
typedef enum 
{
	/* Common Events */		/* Sender   */ /* Receiver */
                                                                
	ECTP_REQ_CONN_CREATE,	/* TO       */ /*    LO LE */
	ECTP_CON_CONN_CREATE,	/*    LO LE */ /* TO       */
                                                         
	ECTP_REQ_LATE_JOIN,		/*    LO LE */ /* TO       */
	ECTP_CON_LATE_JOIN,		/* TO       */ /*    LO LE */
	ECTP_DNY_LATE_JOIN,		/* TO       */ /*    LO LE */
	ECTP_EXP_LATE_JOIN,		/*    LO LE */ /*    LO LE */ 
							/* exceed maximum RXT counter for JR. */
                                                         
	ECTP_REQ_TREE_JOIN,		/*       LE */ /* TO LO    */
	ECTP_CON_TREE_JOIN,		/* TO LO    */ /*       LE */
	ECTP_DNY_TREE_JOIN,		/* TO LO    */ /*       LE */
	ECTP_EXP_TREE_JOIN,		/*       LE */ /*       LE */
							/* exceed maximum RXT counter for TJ. */
	                                                     
	ECTP_REQ_TOKEN_GET,		/* TO    LE */ /* TO    LE */
	ECTP_CON_TOKEN_GET,		/* TO    LE */ /* TO    LE */
	ECTP_DNY_TOKEN_GET,		/* TO    LE */ /* TO    LE */
	ECTP_EXP_TOKEN_GET,		/* TO    LE */ /* TO    LE */
							/* exceed maximum RXT counter for TGR. */
                                                         
	/* Rejection for Token Return does not exist. */
	ECTP_REQ_TOKEN_RET,		/* TO    LE */ /* TO    LE */
	ECTP_CON_TOKEN_RET,		/* TO    LE */ /* TO    LE */
	ECTP_EXP_TOKEN_RET,		/* TO    LE */ /* TO    LE */
							/* exceed maximum RXT counter for TRR. */
                                                         
	ECTP_REQ_CONN_TERM,		/*    LO LE */ /*    LO LE */
	ECTP_REQ_USER_LEAVE,	/* TO LO LE */ /* TO LO LE */


	/* Other Internal Events */

	ECTP_DATA_IO,
	ECTP_DATA_STABLE,		/* All children received the DATA.
							   If DATA packet in send buffer is stable,
							   it must be released out. */

	ECTP_CCT_EXPIRED,		/* When Connection Creation Timer expires,
							   TO enters ESTABLISHED state. */
	ECTP_CWT_EXPIRED,		/* When Connection Wait Timer expires,
							   LE starts Late Join procedure. */

	/* Urgent Events */

	ECTP_URG_CONN_DOWN,		/* Connection is down by some reason. */
	ECTP_URG_SHUTDOWN,		/* Upper layer service requests System shutdown. */

} ectp_event_t ;

/**
 * ECTP-3 Event Item
 */
struct ectp_eventitem {
	
	struct ectp_eventitem *next;
	struct ectp_eventitem *prev;
	struct ectp_eventq *q;

	ectp_event_t type;
	ectp_addr_t from;

	/* If [ packetlen = 0 ] then packet must be NULL.
	 * Otherwise, data will point the received ECTP-3 packet.
	 * 
	 * NOTE:
	 * 		If the event is one of {JR,TJ,TGR,LR,CT}, data field is
	 *		useless.
	 *		Otherwise, data field must point one of {JC,TC,TGC} packet or
	 *		equivalent information.
	 */
	__u32	packetlen;
	void	*packet;
};

#define ECTP_EVENTITEM_LEN	sizeof(struct ectp_eventitem)

/* event.c ----------------------------------------------------------------- */

extern struct ectp_eventitem *ectp_create_event( ectp_event_t type,
		ectp_addr_t *from, size_t fromlen, void *packet, __u32 packetlen );

extern void ectp_destroy_event( struct ectp_eventitem *e );

/**
 * On success a pointer of a new ECTP-3 event item is returned.
 * Otherwise NULL is returned.
 */
static inline struct ectp_eventitem *ectp_ei_alloc()
{
	struct ectp_eventitem *_new = NULL;
	_new = (struct ectp_eventitem *) calloc(1, ECTP_EVENTITEM_LEN);
	return _new;
}

static inline void ectp_ei_free(struct ectp_eventitem *ei)
{
	/* If the given event item has the data malloced, it must be freed. */
	if ( ei->packetlen )
		free(ei->packet);

	free(ei);
}

static inline void ectp_ei_add( struct ectp_eventitem *_new,
								struct ectp_eventitem *prev,
								struct ectp_eventitem *next )
{
	next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

/* 
 * NOTE:
 * 	Before this function is called, MUST check if the entry is NULL.
 */
static inline void ectp_ei_del( struct ectp_eventitem *entry )
{
	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}



/**
 * ECTP-3 Event Queue structure
 */
struct ectp_eventq {

	struct ectp_eventitem *head;
	struct ectp_eventitem *tail;

	/* Any function that manipulates the event queue must turns on this mutex.
	 */
	ectp_mutex_t mutex;

	/* This is a access control semaphore.
	 * It contains the number of available resources (event items).
	 * Initial value is always zero.
	 *
	 * If there is no item in the queue, corresponding event handler will
	 * suspends.
	 */
	ectp_sem_t sem;

	/* PROVIDER
	 * 		Procedure of functions that adds the dispatched event item to
	 * 		the event queue:
	 * 		ectp_event_enq(q, _new);
	 *		  (	ectp_mutex_on(&q->mutex); // internal ops. of adding ft.
	 *			ectp_ei_add();
	 *			ectp_mutex_off(&q->mutex);
	 *			ectp_sem_post(&sem); )
	 * 		Additional operations.
	 * 		
	 *
	 * CONSUMER
	 * 		Procedure of the event handler: 
	 * 		ectp_sem_wait(&sem) / ectp_sem_trywait(&sem);
	 * 		  (	There is at least one items in the queue. )
	 * 		entry = ectp_event_deq();
	 *		  (	ectp_mutex_on(&q->mutex);
	 *		  	entry = ectp_ei_del();
	 *		  	ectp_mutex_off(&q->mutex);
	 *		  	return entry; )
	 */

	/* The number of events in the queue.
	 *
	 * NOTE:
	 * 	"num = resource number of sem"
	 */
	__u32 num;
};

/* event.c ----------------------------------------------------------------- */

/**
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_event_initq( struct ectp_eventq *q );

/**
 * RETURN:
 * 		If empty, 1 is returned.
 * 		Otherwise 0 is returned.
 */
static inline int ectp_event_emptyq( struct ectp_eventq *q )
{
	return (q->num == 0);
}

/**
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_event_enq( struct ectp_eventq *q, struct ectp_eventitem *_new );

/**
 * Blocking functions.
 */
extern struct ectp_eventitem *ectp_event_deq( struct ectp_eventq *q );




/* ------------------------------------------------------------------------- *
 * Definitions for ECTP-3 Timer Event (Alarm)
 * ------------------------------------------------------------------------- */

/**
 * void ectp_timeradd( struct timeval *a,		//
 * 				  struct timeval *b,		// 
 * 				  struct timeval *result	// 
 */
#define ectp_timeradd(a, b, result)					\
do {													\
	(result)->tv_sec  = (a)->tv_sec + (b)->tv_sec;		\
	(result)->tv_usec = (a)->tv_usec + (b)->tv_usec;	\
	if ( (result)->tv_usec >= 1000000 )	{				\
		(result)->tv_sec++;								\
		(result)->tv_usec -= 1000000;					\
	}													\
} while ( 0 )

/**
 * void ectp_timersub( struct timeval *a,       // end time
 *                struct timeval *b,       // start time
 *                struct timebal *result)  // difference
 */
#define ectp_timersub(a, b, result)                     \
do {                                                    \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;       \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;    \
    if ( (result)->tv_usec < 0 ) {                      \
        (result)->tv_sec--;                             \
        (result)->tv_usec += 1000000;                   \
    }                                                   \
} while ( 0 )

typedef enum
{
	ECTP_TIMER_CCT,		/* Connection Creation Timer */
	ECTP_TIMER_CWT,		/* Connection Wating Timer */
	//ECTP_TIMRE_DIT,		/* Data session Invalid Timer */
	ECTP_TIMER_HGT,		/* Heartbeat Generation Timer */
	ECTP_TIMER_RXT		/* Retransmission Timer */
} ectp_timer_t;

struct ectp_timeritem
{
	struct ectp_timeritem	*next;
	struct ectp_timeritem	*prev;
	struct ectp_timerlist	*list;

	ectp_timer_t			type;			/* Timer Type */
	__u32					update_time;	/* Seconds for update timer */
	struct timeval			action_time;	/* Start Time + Timeout value */
	struct ectp_skbuff		*skb;			
};

#define	ECTP_TIMERITEM_LEN	sizeof(struct ectp_timeritem)

static inline struct ectp_timeritem *ectp_ti_alloc()
{
	struct ectp_timeritem *_new = NULL;
	_new = (struct ectp_timeritem *) calloc(1, ECTP_TIMERITEM_LEN);
	return _new;
}

extern void ectp_ti_free(struct ectp_timeritem *alarm);

static inline void ectp_ti_add( struct ectp_timeritem *_new,
								struct ectp_timeritem *prev,
								struct ectp_timeritem *next )
{
	next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

/** Recalculate an action time of the given alarm timer */
static inline int ectp_ti_update( struct ectp_timeritem *alarm )
{
	struct timeval ct, inc;

	if ( !alarm )
		err_ret("ectp_ti_recalc() : given timer is NULL.", -1);

	if ( gettimeofday(&ct, (struct timezone *) NULL) < 0 )
		err_ret("ectp_ti_recalc() : gettimeofday() failed.", -1);

	inc.tv_sec = alarm->update_time;
	inc.tv_usec = 0;

	ectp_timeradd(&ct, &inc, &alarm->action_time);

	return 0;
}

/* Unlink.
 *
 * NOTE:
 * 	Before this function is called, MUST check if the entry is NULL.
 */
static inline void ectp_ti_unlink( struct ectp_timeritem *entry )
{
	if ( !entry )
		return;

	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}

/**
 * time [sec]
 * If [time = 0], default of type will be used.
 */
static inline struct ectp_timeritem *ectp_create_timer(
		ectp_timer_t type, int time, struct ectp_skbuff *skb )
{
	struct ectp_timeritem *_new = NULL;

	if ( !(_new = ectp_ti_alloc()) )
		err_ret("ectp_create_timer() : ectp_ti_alloc() failed.", NULL);

	_new->type = type;
	_new->update_time = time;
	_new->skb = skb;

	if ( ectp_ti_update(_new) < 0 ) {
		ectp_ti_free(_new);
		_new = NULL;
		err_ret("ectp_create_timer() : ectp_ti_update() failed.", NULL);
	}

	return _new;
}

static inline void ectp_destroy_timer( struct ectp_timeritem *alarm )
{
	ectp_ti_free(alarm);
}



struct ectp_timerlist
{
	struct ectp_timeritem *head;
	struct ectp_timeritem *tail;

	ectp_mutex_t mutex;
	__u32 num;
};

/* timer.c ----------------------------------------------------------------- */

/**
 * RETURN:
 * 		On success, 0 is returned.
 * 		Othrewise, -1 is returned.
 */
extern int ectp_timerlist_init( struct ectp_timerlist *list );

static inline int ectp_timerlist_empty( struct ectp_timerlist *list )
{
	return !list->num;
}

/** sorted add */
extern int ectp_timerlist_add( struct ectp_timerlist *list,
		struct ectp_timeritem *_new );

/**
 * If alarm is NULL, the head is deleted from the list and returned.
 * If alarm is specified timer item, it is deleted from the list.
 */
extern struct ectp_timeritem *ectp_timerlist_del(
		struct ectp_timerlist *list, struct ectp_timeritem *alarm );

/*  NOTE: DO NOT RELEASE PEEK ITEM MANUALLY. */

static inline struct ectp_timeritem *ectp_timerlist_peek(
										struct ectp_timerlist *list )
{
	return list->head;
}

/**
 * On success, the pointer of timeval{} is returned.
 * Otherwise, NULL is returned. (No timer / Error)
 */
static inline struct timeval *ectp_timerlist_peek_timeout(
		struct ectp_timerlist *list )
{
	static struct timeval result;
	struct timeval current;

	if ( ectp_timerlist_empty(list) )
		return NULL;

	if ( gettimeofday(&current, (struct timezone *) NULL) < 0 )
		err_ret("ectp_timerlist_peek_timeout() : "
				"gettimeofday() failed.", NULL);

	ectp_timersub(&list->head->action_time, &current, &result);

	if ( result.tv_sec < 0 ) {	/* a timer expired */
		result.tv_sec = 0;
		result.tv_usec = 0;
	}

	return &result;
}

#ifdef	__cplusplus
}
#endif

#endif	/* _ECTP_EVENT_H */
