/* Enhanced Communication Transport Protocol Part-3 Implementation
 *
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * ectp_protocol.h
 * This header represents the ECTP-3 protocol specification.
 *
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers	<saintpaul1978@mail.knu.ac.kr>
 *						<knucsid@gmail.com>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 *	Jae Sung Park				<knucsid@gmail.com>
 *
 */

#include <ectp/types.h>

#ifndef	_ECTP_PROTOCOL_H
#define	_ECTP_PROTOCOL_H



#ifdef	__cplusplus
extern "C" {
#endif



/* ------------------------------------------------------------------------- *
 * Packet Types for ECTP-3 Specification
 * ------------------------------------------------------------------------- */

/**
 * 8.1 Base Header
 */
struct ectp_basehdr
{
	__u8	next:4,			/* Next Element */
			version:2,		/* ECTP-3 Protocol Version */
			ct:2;			/* Connection Type */
	__u8	packet_type;	/* Packet Type */
	__u16	checksum;		/* Adler32 Checksum */
	__u32	conn_id;		/* Connection ID */
	__u32	psn;			/* Packet Sequence Number (PSN) */
	__u16	payload_length;	/* Payload Length */
	__u8	f:1,			/* F, a flag bit */
			reserved:7;
	__u8	token_id;		/* Token ID */
};

#define	ECTP_BASEHDR_LEN			sizeof(struct ectp_basehdr)

/**
 * 8.2 Extention Elements (4bits)
 */
#define	ECTP_EE_NOELEMENT		0x00
#define	ECTP_EE_CONNECTION		0x01
#define	ECTP_EE_ACKNOLEDGEMENT	0x02
#define	ECTP_EE_MEMBERSHIP		0x03
#define ECTP_EE_TIMESTAMP		0x04
/*
 * The encoding value of '0101' is reserved for the QoS extention element,
 * which is used in ECTP-4.
 */
#define	ECTP_EE_ADDRESS			0x06


/**
 * 8.2.1 Connection Element
 */
struct ectp_ee_conn
{
	__u8	next:4,				/* Next Element */
			tco:2,				/* TCO (Tree Configuration Option) */
#define	ECTP_TCO_LEVEL1			0x00
#define	ECTP_TCO_LEVEL2			0x01
#define	ECTP_TCO_GENERAL		0x02
#define	ECTP_TCO_RESERVED		0x03
			r:2;				/* R (Reserved) */
	__u8	agn;				/* AGN (ACK Generation Number) */
	__u16	mss;				/* MSS (Maximum Segment Size) */
};

#define	ECTP_EE_CONNECTION_LEN	sizeof(struct ectp_ee_conn)


/**
 * 8.2.2 Acknowledgement Element
 */
struct ectp_ee_ack
{
	__u8	next:4,				/* Next Element */
			bitmap_length:4;	/* Bitmap Length in unit of word (4 bytes) */
	__u8	valid_bitmap;		/* Valid Bitmap Length */
	__u16	reserved;

	/* bitmap_length words ACK Bitmap here. */
	//__u32	*ack_bitmap;		/* ACK Bitmap with a variable length */
};


/**
 * 8.2.3 Membership Element
 */
struct ectp_ee_member
{
	__u8	next:4,				/* Next Element */
			tree_level:4;		/* Tree Level */
	__u8	child_id;			/* Child ID */
	__u16	adn;				/* ADN (Active Descendant Number) */
};

#define	ECTP_EE_MEMBERSHIP_LEN	sizeof(struct ectp_ee_member)


/**
 * 8.2.4 Timestamp Element
 */
struct ectp_ee_timestamp
{
	__u32	next:4,				/* Next Element */
			reserved:28;
	__u32	sec;				/* Timestamp in seconds */
	__u32	usec;				/* Timestamp in microseconds */

};

/**
 * 8.2.5 Address Element
 */
struct ectp_ee_addr
{
	__u32	next:4,				/* Next Element */
			ip_ver:4,			/* IP Version */
			reserved:24;

	/* Here comes 32 bits or 128 bits IP Address according to ip_ver. */
	/* __u8 addr[4]; */
	/* __u8 addr[16] */
};


/**
 * 8.3 ECTP-3 Packet Format
 */
typedef enum
{
	ECTP_PKT_CR 	= 0x01,
#define	ECTP_PKT_CR		 ECTP_PKT_CR
	ECTP_PKT_CC		= 0x02,
#define ECTP_PKT_CC		 ECTP_PKT_CC
	ECTP_PKT_TJ		= 0x03,
#define ECTP_PKT_TJ		 ECTP_PKT_TJ
	ECTP_PKT_TC		= 0x04,
#define ECTP_PKT_TC		 ECTP_PKT_TC
	ECTP_PKT_DT		= 0x05,
#define ECTP_PKT_DT		 ECTP_PKT_DT
	ECTP_PKT_ND		= 0x06,
#define ECTP_PKT_ND		 ECTP_PKT_ND
	ECTP_PKT_RD		= 0x07,
#define ECTP_PKT_RD		 ECTP_PKT_RD
	ECTP_PKT_ACK	= 0x08,
#define ECTP_PKT_ACK	 ECTP_PKT_ACK
	ECTP_PKT_HB		= 0x09,
#define ECTP_PKT_HB		 ECTP_PKT_HB
	ECTP_PKT_HBACK	= 0x0E,
#define ECTP_PKT_HBACK	 ECTP_PKT_HBACK
	ECTP_PKT_JR		= 0x0A,
#define ECTP_PKT_JR		 ECTP_PKT_JR
	ECTP_PKT_JC		= 0x0B,
#define ECTP_PKT_JC		 ECTP_PKT_JC
	ECTP_PKT_LR		= 0x0C,
#define ECTP_PKT_LR		 ECTP_PKT_LR
	ECTP_PKT_CT		= 0x0D,
#define ECTP_PKT_CT		 ECTP_PKT_CT
	ECTP_PKT_TGR	= 0x11,
#define ECTP_PKT_TGR	 ECTP_PKT_TGR
	ECTP_PKT_TGC	= 0x12,
#define ECTP_PKT_TGC	 ECTP_PKT_TGC
	ECTP_PKT_TRR	= 0x13,
#define ECTP_PKT_TRR	 ECTP_PKT_TRR
	ECTP_PKT_TRC	= 0x14
#define ECTP_PKT_TRC	 ECTP_PKT_TRC
} ectp_packet_t;

/**
 * 8.3.1 Connection Creation Request (CR)
 */
typedef struct {
	struct ectp_basehdr				basehdr;
	struct ectp_ee_conn			connection;
} ectp_packet_ccr;

/**
 * 8.3.2 Connection Creation Confirm (CC)
 */
typedef struct ectp_basehdr		ectp_packet_ccc;

/**
 * 8.3.3  Tree Join Request (TJ)
 */
typedef struct ectp_basehdr		ectp_pkt_tj;
#define	ECTP_PKT_TJ_LEN		sizeof(ectp_pkt_tj)

/**
 * 8.3.4 Tree Join Confirm (TC)
 */
typedef struct {
	struct ectp_basehdr			basehdr;
	struct ectp_ee_member		membership;
	/* Optional field address element is NOT used in this implementation. */
	/* struct ectp_ee_addr			address; */
} ectp_pkt_tc;

#define	ECTP_PKT_TC_LEN		sizeof(ectp_pkt_tc)

/**
 * 8.3.5 Data (DT)
 */
typedef struct ectp_basehdr		ectp_pkt_dt;

/* After the base header  user data comes. */

#define ECTP_PKT_DT_LEN		sizeof(ectp_pkt_dt)

/**
 * 8.3.6 Null Data (ND) is no more necessary.
 */
typedef struct ectp_basehdr		ectp_pkt_nd;

/**
 * 8.3.7 Retransmission Data (RD)
 */
typedef ectp_pkt_dt	ectp_pkt_rd;

/**
 * 8.3.8 Acknowledgement (ACK)
 */
typedef struct {
	struct ectp_basehdr				basehdr;
	struct ectp_ee_member		membership;
	struct ectp_ee_ack			acknowledgement;

	/* acknowledgement.bitmap_length words ACK Bitmap here. */
} ectp_pkt_ack;
#define ECTP_PKT_ACK_LEN	sizeof(ectp_pkt_ack)

/**
 * 8.3.9 Heartbeat (HB)
 */
typedef struct {
	struct ectp_basehdr				basehdr;
	struct ectp_ee_member		membership;
	/* struct ectp_ee_timestamp	timestamp;	// option */
} ectp_pkt_hb;

/** 
 * 8.3.10 Heartbeat Acknowledgement (HBACK)
 */
typedef struct {
	struct ectp_basehdr				basehdr;
	struct ectp_ee_member		membership;
	struct ectp_ee_ack			acknowledgement;
	/* struct ectp_ee_timestamp	timestamp;	// option */
} ectp_pkt_hback;

/**
 * 8.3.11 Late Join Request (JR)
 */
typedef struct ectp_basehdr		ectp_pkt_jr;

#define	ECTP_PKT_JR_LEN		sizeof(ectp_pkt_jr)

/**
 * 8.3.12 Late Join Confirm (JC)
 */
typedef struct {
	struct ectp_basehdr			basehdr;
	struct ectp_ee_conn			connection;

 	/* Optional field address element is NOT used in this implementation. */
	/* struct ectp_ee_addr			addr; */
} ectp_pkt_jc;

#define ECTP_PKT_JC_LEN		sizeof(ectp_pkt_jc)

/**
 * 8.3.13 User Leave Request (LR)
 */
typedef struct ectp_basehdr		ectp_pkt_lr;

#define	ECTP_PKT_LR_LEN		sizeof(ectp_pkt_lr)

/**
 * 8.3.14 Connection Termination Request (CT)
 */
typedef struct ectp_basehdr		ectp_pkt_ct;

#define	ECTP_PKT_CT_LEN		sizeof(ectp_pkt_ct)

/**
 * 8.3.15 Token Get Request (TGR)
 */
typedef struct ectp_basehdr		ectp_pkt_tgr;

#define	ECTP_PKT_TGR_LEN		sizeof(ectp_pkt_tgr)

/**
 * 8.3.16 Token Get Confirm (TGC)
 */
typedef struct ectp_basehdr		ectp_pkt_tgc;

#define	ECTP_PKT_TGC_LEN		sizeof(ectp_pkt_tgc)

/**
 * 8.3.17 Token Return Request (TRR)
 */
typedef struct ectp_basehdr		ectp_pkt_trr;

#define	ECTP_PKT_TRR_LEN		sizeof(ectp_pkt_trr)

/**
 * 8.3.18 Token Return Confirm (TRC)
 */
typedef struct ectp_basehdr		ectp_pkt_trc;

#define	ECTP_PKT_TRC_LEN		sizeof(ectp_pkt_trc)


#ifdef	__cplusplus
}
#endif

#endif	/* _ECTP_PROTOCOL_H */
