/* Enhanced Communication Transport Protocol Part-3 Implementation
 *
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * This header represents the ECTP-3 socket structure and protocol control
 * block.
 *
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers	<saintpaul1978@mail.knu.ac.kr>
 *						<knucsid@gmail.com>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 *	Jae Sung Park				<knucsid@gmail.com>
 *
 */

#include <ectp/types.h>
#include <ectp/wrapper.h>
#include <ectp/event.h>

#include <stdlib.h>			/* for malloc()/free() */


#ifndef	_ECTP_STRUCTS_H
#define	_ECTP_STRUCTS_H

#ifdef	__cplusplus
extern "C" {
#endif



/* ------------------------------------------------------------------------- *
 * Forward Declarations for Data Structures
 * ------------------------------------------------------------------------- */

struct ectp_globals;
struct ectp_socket;
struct ectp_endpoint;
struct ectp_connection;
struct ectp_tree;
struct ectp_tokenset;
struct ectp_childset;
struct ectp_skbuff;
struct ectp_skblist;


/* ------------------------------------------------------------------------- *
 * ECTP-3 Global Structure
 * ------------------------------------------------------------------------- */

/**
 * Maximum number of the ectp socket
 */
#define	ECTP_MAX_SOCKETS	8

/**
 * ECTP-3 Global Structure
 *
 * Currently global structure contains only socket list information.
 */
struct ectp_globals {

	/** the current number of the ectp socket */
	__u8				nsock;

	/**
	 * Mutex for ECTP-3 Socket
	 * - Mutex about syncronization of Access multiple ECTP-3 sockets.
	 * - It is needed to initialize this mutex.
	 *
	 * NOTE: Currently it is not important.
	 */
	ectp_mutex_t		mutex;

	/**
	 * The Global ECTP-3 Socket Array Pointer.
	 * Each index represents ECTP-3 Socket Descriptor.
	 */
	struct ectp_socket	*socks[ECTP_MAX_SOCKETS];

};

/**
 * ECTP-3 Global Variables
 */
extern struct ectp_globals *ectp_global;

/**
 * Socket Pointer that is used to Current Process.
 */
extern struct ectp_socket *ectp_current;

/* socket.c ---------------------------------------------------------------- */

/**
 * ectp_create_globals() create the global variable structure.
 * - On success a pointer to the allocated memory is returned.
 * - Otherwise NULL is returned.
 */
extern struct ectp_globals *ectp_create_globals();

extern void ectp_destroy_globals( struct ectp_globals *global );


/* -------------------------------------------------------------------------- *
 * ECTP-3 Token Set Structure
 * -------------------------------------------------------------------------- */

/**
 * ECTP-3 Token Set
 */

struct ectp_tokenset {

	/* The n represents total number of tokens that are currently assigned to
	 * TU.
	 */
	__u8			n;

	/* bitmap is a bit array which contains 255 valid tokens. Each bit-position
	 * represents corresponding token number assigned to TU.
	 */

	/* NOTE:
	 * Because the Token ID '0' is reserved for use of TO, we do not use:
	 * 		LSB(2^0 bit) of bitmap. It is always zero.
	 * 		su_addrs[0]. It is always NULL.
	 */

	__u32			bitmap[ECTP_NTOKEN_WORDS];	/* bitmap[8] */
	ectp_addr_t		*su_addrs[ECTP_NTOKEN];		/* su_addrs[256] */
};

#define ECTP_TOKENSET_LEN	sizeof(struct ectp_tokenset)

/* ecto_control.c ---------------------------------------------------------- */

extern struct ectp_tokenset *ectp_create_tokenset();

extern int ectp_destroy_tokenset( struct ectp_tokenset *set );

static inline int ectp_token_isvalid( ectp_token_t token )
{
	return IS_TOKEN_VALID(token);
}

static inline int ectp_token_ismember( struct ectp_tokenset *set )
{
	return 0;
}

/**
 * Find First Set-bit in the ECTP-3 Token Map except for LSB.
 * 
 * This function  returns  the position of the first bit set, in the form of
 * 2's power, or -1 if no bits are set in set->bitmap.
 *
 * - 	If bitmap = (11111110), pos = 1;
 * 	                       ^  
 * - 	If bitmap = (11111100), pos = 2;
 *   	                  ^   
 * - 	If bitmap = (10000000), pos = 7;
 * 	                 ^        
 * - 	If bitmap = (00000000), pos = -1;
 *
 * NOTE:
 * -	This function never return 0, because Token ID 0 is reserved for TO.
 *  	Consider the LSB is always clear.
 */
static inline int ectp_token_ffs( struct ectp_tokenset *set )
{
	int		wc  = 0;	/* Word counter */
	int		pos = 0;	/* array position(index) */
	int		tmp = 0;
	int 	*bitmap = NULL;

	/* Clear LSB because it is reserved to TO. */
	set->bitmap[0] &= ~1UL;
	bitmap = (int *) set->bitmap;

	/* Lower boundary condition! */
	if ( (pos = ffs(*bitmap)) )
		goto tokenffs_out;

	bitmap++;
	wc++;

	/* Upper boundary condition! */
	while ( !(tmp = ffs(*bitmap)) && wc < ECTP_NTOKEN_WORDS ) {
		pos = tmp;
		bitmap++;
		wc++;
	}

	/* Upper boundary condition! */
	if ( wc < ECTP_NTOKEN_WORDS )
		pos = tmp;

tokenffs_out:
	return ( !pos ? -1 : (--pos + wc * ECTP_NTOKEN_BPW));
}
	
/**
 * Find First Zero-bit in ECTP-3 Token Map except for LSB.
 *
 * - 	If bitmap = (00000000), pos = 1;
 * 	                       ^  
 * - 	If bitmap = (00000010), pos = 2;
 * 	                      ^   
 * - 	If bitmap = (01111110), pos = 7;
 * 	                 ^        
 * - 	If bitmap = (11111110), pos = -1;	----> bitmap is full.
 *
 * NOTE:
 * -	This function never return 0.
 * 		Consider the LSB is always set.
 *
 * RETURN:
 */
static inline int ectp_token_ffz( struct ectp_tokenset *set )
{
	int		wc  = 0;	/* Word counter */
	int		pos = 0;	/* array position(index) */
	int		tmp = 0;
	int 	*bitmap = NULL;

	/* Set LSB because it is reserved to TO. */
	set->bitmap[0] |= 1UL;
	bitmap = (int *) set->bitmap;

	/* Lower boundary condition! */
	if ( (pos = ffs(~*bitmap)) )
		goto tokenffz_out;

	bitmap++;
	wc++;

	while ( !(tmp = ffs(~*bitmap)) && wc < ECTP_NTOKEN_WORDS ) {
		pos = tmp;
		bitmap++;
		wc++;
	}

	/* Upper boundary condition! */
	if ( wc < ECTP_NTOKEN_WORDS )
		pos = tmp;

tokenffz_out:
	/* Clear LSB */
	set->bitmap[0] &= ~1UL;

	return ( !pos ? -1 : (--pos + wc * ECTP_NTOKEN_BPW));
}

static inline void ectp_token_emptyset( struct ectp_tokenset *set )
{
	int	wc = 0;

	for ( wc = 0; wc < ECTP_NTOKEN_WORDS; wc++ )
		set->bitmap[wc] &= 0UL;

	for ( wc = 0; wc < ECTP_NTOKEN; wc++ )
		set->su_addrs[wc] = NULL;

	set->n = 0;
}

static inline void ectp_token_fillset( struct ectp_tokenset *set )
{
	int	wc = 0;

	for ( wc = 0; wc < ECTP_NTOKEN_WORDS; wc++ )
		set->bitmap[wc] |= ~0UL;

	/* Because the Token ID '0' is reserved for use of TO. */
	set->bitmap[0] &= ~1UL;
	set->n = ECTP_NTOKEN - 1;
}

/**
 * RETURN:
 * 		On success, zero is returned.
 */
static inline void ectp_token_addset( struct ectp_tokenset *set,
									  int token, ectp_addr_t *su )
{
	set->bitmap[ token / ECTP_NTOKEN_BPW ] |= 
		1UL << ( token % ECTP_NTOKEN_BPW );
	if ( su ) {
		set->su_addrs[token] = (ectp_addr_t*)malloc(sizeof(ectp_addr_t));
		memcpy( set->su_addrs[token], su, sizeof(ectp_addr_t));
	}
	set->n++;
}

static inline void ectp_token_delset( struct ectp_tokenset *set, int token )
{
	set->bitmap[ token / ECTP_NTOKEN_BPW ] &=
		~(1UL << (token % ECTP_NTOKEN_BPW));

	if ( set->su_addrs[token] != NULL ) {
		free(set->su_addrs[token]);
		set->su_addrs[token] = NULL;
	}
	set->n--;
}



/* -------------------------------------------------------------------------- *
 * ECTP-3 Child Set Structure
 * -------------------------------------------------------------------------- */

struct ectp_childset {

	/* The n represents total number of children that are currently active.
	 *
	 * NOTE:
	 * 		n is different to adn. adn is the overall decendent number.
	 * 		For instance, adn includes grand-children but n does not. In other
	 * 		words n includes only directly connected children.
	 */
	__u8			n;

	/* bitmap is a bit array which contains 256 valid Child ID. Each
	 * bit-position represents corresponding Child ID assigned to Child.
	 */

	/* NOTE:
	 * Because the Child ID of '0' is used to indicate that this field must be
	 * ignored, we do not use:
	 * 		LSB(2^0 bit) of bitmap. It is always zero.
	 * 		child_addrs[0]. It is always NULL.
	 */

	__u32			bitmap[ECTP_NCID_WORDS];		/* bitmap[8] */
	ectp_addr_t		*child_addrs[ECTP_NCID];		/* child_addrs[256] */
};

#define	ECTP_CHILDSET_LEN	sizeof(struct ectp_childset)


struct ectp_childset *ectp_create_childset();

extern int ectp_destroy_childset( struct ectp_childset *set );

static inline int ectp_child_ismember( struct ectp_childset *set )
{
	return 0;
}

/**
 * Find First Set-bit in the ECTP-3 Child Map except for LSB.
 *
 * This function  returns  the position of the first bit set, in the form of
 * 2's power, or -1 if no bits are set in set->bitmap.
 *
 * - 	If bitmap = (11111110), pos = 1;
 * 	                       ^  
 * - 	If bitmap = (11111100), pos = 2;
 *   	                  ^   
 * - 	If bitmap = (10000000), pos = 7;
 * 	                 ^        
 * - 	If bitmap = (00000000), pos = -1;
 *
 * NOTE:
 * - 	This function never return 0, because Child ID 0 will be ignored.
 *   	Consider the LSB is always clear.
 */
static inline int ectp_child_ffs( struct ectp_childset *set )
{
	int		wc  = 0;	/* Word counter */
	int		pos = 0;	/* array position(index) */
	int		tmp = 0;
	int 	*bitmap = NULL;

	/* Clear LSB because will be ignored. */
	set->bitmap[0] &= ~1UL;
	bitmap = (int *) set->bitmap;

	/* Lower boundary condition! */
	if ( (pos = ffs(*bitmap)) )
		goto childffs_out;

	bitmap++;
	wc++;

	/* Upper boundary condition! */
	while ( !(tmp = ffs(*bitmap)) && wc < ECTP_NCID_WORDS ) {
		pos = tmp;
		bitmap++;
		wc++;
	}

	/* Upper boundary condition! */
	if ( wc < ECTP_NCID_WORDS )
		pos = tmp;

childffs_out:
	return ( !pos ? -1 : (--pos + wc * ECTP_NCID_BPW));
}

/**
 * Find First Zero-bit in ECTP-3 Child Map except for LSB.
 *
 * - 	If bitmap = (00000000), pos = 1;
 * 	                       ^  
 * - 	If bitmap = (00000010), pos = 2;
 * 	                      ^   
 * - 	If bitmap = (01111110), pos = 7;
 * 	                 ^        
 * - 	If bitmap = (11111110), pos = -1;	----> bitmap is full.
 *
 * NOTE:
 * -	This function never return 0.
 * 		Consider the LSB is always set.
 *
 * RETURN:
 */
static inline int ectp_child_ffz( struct ectp_childset *set )
{
	int		wc  = 0;	/* Word counter */
	int		pos = 0;	/* array position(index) */
	int		tmp = 0;
	int 	*bitmap = NULL;

	/* Set LSB because it is reserved to TO. */
	set->bitmap[0] |= 1UL;
	bitmap = (int *) set->bitmap;

	/* Lower boundary condition! */
	if ( (pos = ffs(~*bitmap)) )
		goto childffz_out;

	bitmap++;
	wc++;

	while ( !(tmp = ffs(~*bitmap)) && wc < ECTP_NCID_WORDS ) {
		pos = tmp;
		bitmap++;
		wc++;
	}

	/* Upper boundary condition! */
	if ( wc < ECTP_NCID_WORDS )
		pos = tmp;

childffz_out:
	/* Clear LSB */
	set->bitmap[0] &= ~1UL;

	if ( pos )
		set->n++;

	return ( !pos ? -1 : (--pos + wc * ECTP_NCID_BPW));
}

static inline void ectp_child_emptyset( struct ectp_childset *set )
{
	int	wc = 0;
	while( wc < ECTP_NCID_WORDS ) {
		set->bitmap[wc] &= 0UL;
		wc++;
	}

	wc = 0;
	while( wc < ECTP_NCID ) {
		set->child_addrs[wc] = NULL;
		wc++;
	}

	set->n = 0;
}

static inline void ectp_child_fillset( struct ectp_childset *set )
{
	int	wc = 0;
	while( wc < ECTP_NCID_WORDS ) {
		set->bitmap[wc] |= ~0UL;
		wc++;
	}
	
	set->bitmap[0] &= ~1UL;
	set->n = ECTP_NCID - 1;
}

static inline void ectp_child_addset( struct ectp_childset *set,
									  int cid, ectp_addr_t *child )
{
	set->bitmap[ cid / ECTP_NCID_BPW ] |=
		1UL << ( cid % ECTP_NCID_BPW );

	if ( child ) {
		set->child_addrs[cid] = (ectp_addr_t*)malloc(sizeof(ectp_addr_t));
		memcpy( set->child_addrs[cid], child, sizeof(ectp_addr_t));
	}
}

static inline void ectp_child_delset( struct ectp_childset *set, int cid )
{
	set->bitmap[ cid / ECTP_NCID_BPW ] &=
		~(1UL << (cid % ECTP_NCID_BPW));

	if ( set->child_addrs[cid] != NULL ) {
		free(set->child_addrs[cid]);
		set->child_addrs[cid] = NULL;
	}
	set->n--;
}



/* ------------------------------------------------------------------------- *
 * ECTP-3 Socket Buffer Structure
 * ------------------------------------------------------------------------- */

/**
 * ECTP-3 socket buffer structure
 */
struct ectp_skbuff {

	struct ectp_skbuff *next;
	struct ectp_skbuff *prev;

	/* The pointer to the parent socket buffer list */
	struct ectp_skblist *list;

	/* Internal UDP socket descriptor used or to use */
	__u32 sd;

	/* The address of data sender/receiver */
	ectp_addr_t addr;

	/* The length of addr */
	__u32 addrlen;

	/* The length of packet */
	__u32 pktlen;

	/* ECTP-3 packet received or to send */
	void *pkt;

	/* PSN of the received packet */
	__u32 psn;

	/* 9.3.2.3 Retransmissions and ACK Aggregation by LO
	 *
	 * After a parent retransmits an RDATA packet, it will ignore any
	 * subsequent retransmission requests for the same packet during the RXT.
	 */
	__u8 has_sent;

	/* Retransmit count */
	__u8 rtxcnt;

	/* Ack count
	 * 		The initial value of ack count is the number of children.
	 * 		It decrease when every corresponding ACK is received as well as it
	 *  	is larger than zero.
	 * 		If ack count is zero, ECTP_EVENT_DATA_STABLE occurs and event
	 * 		handler will remove this socket buffer from the list and deallocate
	 * 		its memory.
	 */
	__u16 ackcnt;
};

#define ECTP_SKBUFF_LEN	sizeof(struct ectp_skbuff)

/**
 * ectp_skb_alloc() allocates a new memory for the ECTP-3 socket buffer.
 *
 * @param	eso		- related ECTP-3 socket
 * @param	sd		- internal UDP socket used or to use
 * @param	addr	- address of the sender(rcvbuf) or receiver(sndbuf)
 *
 * RETURN:
 * 		On success a pointer to a new ECTP-3 socket buffer is returned.
 * 		Otherwise NULL is returned.
 */ 	
//static inline struct ectp_skbuff *ectp_skb_alloc(
static inline struct ectp_skbuff *ectp_skb_alloc(
		struct ectp_socket *eso, int sd, ectp_addr_t *addr, socklen_t addrlen)
{
	struct ectp_skbuff *skb = NULL;

	skb = (struct ectp_skbuff *) calloc(1, ECTP_SKBUFF_LEN);

	if ( !skb ) {
		return NULL;
	}

	skb->sd = sd;
	skb->addrlen = addrlen;
	memcpy(&skb->addr, addr, skb->addrlen);

	/* Followings are filled when corresponding function is called. */

	return skb;
}

/**
 * ectp_skb_free() releases out the memory allocated to skb and skb->packet.
 */
static inline void ectp_skb_free(struct ectp_skbuff *skb)
{
	if ( skb->pktlen )
		free(skb->pkt);

	if ( skb ) {
		free(skb);
		skb = NULL;
	}
}

/**
 * ectp_skb_add() inserts a new entry between two known consecutive entries.
 */
static inline void ectp_skb_add( struct ectp_skbuff *_new,
								 struct ectp_skbuff *prev,
								 struct ectp_skbuff *next)
{
	 next->prev = _new;
	_new->next = next;
	_new->prev = prev;
	prev->next = _new;
}

/**
 * ectp_skb_unlink() cut off links of the designated entry.
 *
 * NOTE:
 * 		This function does not deallocate the memory for the entry.
 * 		If you need it, use ectp_skb_free() after ectp_skb_unlink().
 */
static inline void ectp_skb_unlink( struct ectp_skbuff *entry )
{
	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}



/**
 * ECTP-3 socket buffer list
 *
 * It is used to ECTP-3 send/receive buffer of the socket structure.
 */
struct ectp_skblist {
	
	/* head node */
	struct ectp_skbuff *head;

	/* tail node */
	struct ectp_skbuff *tail;

	/* list access mutex */
	ectp_mutex_t mutex;

	ectp_sem_t sem;

	/* total number of nodes of the ECTP-3 socket buffer list. */
	__u32 num;

	/* total data length in bytes of the ECTP-3 socket buffer list. */
	__u32 length;

	/* Maximum data length in bytes of the ECTP-3 socket buffer.
	 * NOTE:
	 * 		This field represents the maximum length of the data that can be
	 * 		allocated. You should consider additional spaces of structures.
	 * 		In other words, length  = [sizeof(*ectp_skbuff->packet)].
	 * 		Overall allocated memory space is:
	 * 		ectp_skblist->length + ectp_skblist->num * ECTP_SKBUFF_LEN
	 */
	__u32 max_length;
};

/**
 * ectp_skblist_init() makes ready to use ECTP-3 socket buffer as a
 * input/output queue.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_skblist_init( struct ectp_skblist *list, __u32 max_length );

/**
 * ectp_skblist_spare() checks the spare space of the socket buffer list.
 *
 * RETURN:
 * 		The length that can be used is returned.
 */
static inline __u32 ectp_skblist_spare( struct ectp_skblist *list )
{
	return (list->max_length - list->length);
}

/**
 * ectp_skblist_empty() tells us whether the socket buffer list is empty.
 *
 * RETURN:
 * 		If empty, 1 is returned.
 * 		Otherwise 0 is returned.
 */
static inline int ectp_skblist_empty( struct ectp_skblist *list )
{
	return (list->num == 0);
}

/**
 * ectp_skblist_add() appends a new socket buffer to the given list.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_skblist_add( struct ectp_skblist *list,
									 struct ectp_skbuff *_new );

/**
 * ectp_skblist_del() deletes head node and return it.
 * -	If 'skb' is NULL, the head node of 'list' is deleted and returned.
 * -	If 'skb' is specified socket buffer, it is deleted from the list.
 */
extern struct ectp_skbuff *ectp_skblist_del( struct ectp_skblist *list,
											 struct ectp_skbuff  *skb );





/* -------------------------------------------------------------------------- *
 * ECTP-3 Socket Structure
 * -------------------------------------------------------------------------- */

/**
 * struct ectp_socket{}
 *
 * Traditionally a socket is a peer-to-peer communication endpoint that hides
 * the details of the network protocols and the transport mechanisms.
 * 
 * In ECTP-3, a socket is responsible to duplex multicast communications
 * between one TO and many TUs.
 */
struct ectp_socket {

	/* ECTP-3 socket descriptor */
	ectp_sockdes_t	esd;

	/* The UDP socket to send/receive multicast packet. */
	int	m_sd;

	/* The UDP socket to send/receive unicast packet. */
	int u_sd;

	/* Domain */
	int	domain;

	/* In ECTP-3, each packet uses the following types of IP address and port
	 * number for its source and destination address.
	 * 		Group IP address and Local IP address.
	 * 		Group port and Local port. */

	/* Local address */
	ectp_addr_t	local_addr;

	/* Group address */
	ectp_addr_t	group_addr;

	/* ECTP-3 transport endpoint for duplex multicast communication. */
	struct ectp_endpoint *ep;

	/* User Callback */
	struct ectp_callback callback;

	/* ECTP-3 Timer List */
	struct ectp_timerlist timer_list;

	/* ECTP-3 Event Queue:
	 * 		Join Queue  [TO/LO]     - LJR, TJR -+--> "Not used by TU" 
	 * 		Token Queue [TO]        - TGR ------+
	 * 		Event Queue	[TO/LO/TU]  - ULR, CTR, TGC ---> "Used by all"
	 */
	struct ectp_eventq joinq, tokenq, eventq;

	/* ECTP-3 Internal Event Handle Set */
	struct ectp_handleset hset;

	/* Maximum length of the send buffer. */
	__u32 max_sndbuf;

	/* Maximum length of the receive buffer. */
	__u32 max_rcvbuf;
	
	/* Send buffers for multicast/unicast channel.
	 */ 
	struct ectp_skblist m_sndbuf, u_sndbuf;

	/* Receive buffers for multicast/unicast channel. */
	struct ectp_skblist m_rcvbuf, u_rcvbuf;


};

#define ECTP_SOCKET_LEN		sizeof(struct ectp_socket)

/* socket.c ---------------------------------------------------------------- */

/**
 * ectp_create_socket() creates ECTP-3 endpoint for communications
 * - On success, a pointer of a new ECTP-3 socket is returned.
 *   Otherwise, NULL is returned.
 */
extern struct ectp_socket *ectp_create_socket( int domain, ectp_role_t role );

/**
 *
 */
extern int ectp_destroy_socket( struct ectp_socket *eso );

/**
 *
 */
static inline struct ectp_socket *ectp_get_socket( ectp_sockdes_t esd )
{
	return (ectp_current = ectp_global->socks[esd]);
}


/* Following Functions are used to above two functions.
 *
 * RULE:
 * 		ectp_create_socket() / ectp_destroy_socket()
 *       |           |          |            |
 *       +---+-------+----------+            |
 *           |       +---+-------------------+ 
 *           |    +------+
 *           |    |      
 *           v    v
 * 			ectp_sk_alloc();
 * 			ectp_sk_init();
 * 			ectp_sk_free();
 */

/**
 * ectp_sk_alloc() allocates new memory for given ECTP-3 socket pointer.
 *
 * RETURN:
 * 		On success a new pointer to ECTP-3 socket is returned.
 * 		Otherwise NULL is returned.
 */
extern struct ectp_socket *ectp_sk_alloc();

/**
 * ectp_sk_free() deallocates memory of the given ECTP-3 socket.
 */
extern void ectp_sk_free( struct ectp_socket *eso );

/**
 * ectp_sk_init() initializes each field of the given ECTP-3 socket. It also
 * open a new UDP socket for communications.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_sk_init( struct ectp_socket *eso,
						 int domain, ectp_role_t role );

/*
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_sk_bind( struct ectp_socket *eso,
						 SS *local_addr, SS *group_addr );





/* -------------------------------------------------------------------------- *
 * Definitions for ECTP-3 Communication Endpoint, TCB, Tree
 * -------------------------------------------------------------------------- */

/**
 * ECTP-3 Communication Endpoint
 *
 * Endpoint represents the logical sender/receiver of ECTP-3 messages. It
 * contaions all information that is required to communicate with the parent or
 * child in the ECTP-3 communication tree.
 *
 * According to implementation, it is possible to create several endpoints.
 * Each endpoint are linked to the linked list and the virtual or real kernel
 * must manage all the stats of the endpoints created.
 *
 * For simplicity, we doesn't implement the virtual kernel, therefore, some
 * fields required to support several endpoints can be skipped.
 */

struct ectp_endpoint {

	/* A List of Transport Endpoints
	 *
	 */

	/* Backpointer to ECTP-3 socket to which this endpoint beloing. */
	struct ectp_socket *eso;

	/* ECTP-3 Connection Structure List (TCB list) */
	struct ectp_connection *cn;

	/* ECTP-3 Control Tree */
	struct ectp_tree *tree;

	/* ECTP-3 Finite State Machine */
	ectp_state_t state;

	/* A role in ECTP-3 communication tree.
	 * This field is valid only after creating/joining the ECTP-3 connection.
	 * 		ECTP_ROLE_TO	- TO
	 * 		ECTP_ROLE_LE	- TU
	 * 		ECTP_ROLE_LO	- LO
	 */
	ectp_role_t	role;

	/* 7.5 Tokens
	 * In ECTP-3, a token represents the right for a TU to send a unicast data
	 * to TO. Before transmitting the data each TU must get a token from the TO.
	 *
	 * Each token is represented as a 1-byte non-neagtive integer (Token ID).
	 * Token ID is ranged between 1 and 255. The Token ID of '0' is reserved
	 * for use of TO. At the TO side, the Token ID can be used to identify who
	 * is sending the unicast data.
	 *
	 * This field represents the token that are received from TO. If token is
	 * non-negative integer, the local endpoint is SU.
	 *
	 * If TO, 'token' is always '0'.
	 * If LE or LO, 'token' is normally set to ECTP_INVALID_TOKEN.
	 *
	 * SEE:
	 * 		ECTP_TOKEN_TO(0), ECTP_INVALID_TOKEN(-1)
	 */
	ectp_token_t token;

	/* When TU sends TGR, it will be blocked. As well as TU receives TGC, it
	 * will awake. For this operation, token_get_mutex must be initialized with
	 * zero value. Specified procedure is as follows:
	 *
	 * 		Init phase:	ectp_mutex_init(&token_get_mutex);
	 * 					ectp_mutex_on(&token_get_mutex);
	 *
	 * 		TGR phase:	mtoken_get();
	 * 					 ( ectp_mutex_on(&token_get_mutex); ----> block!
	 *					   // TU receives TGC, and event handler call
	 *					   // ectp_mutex_off(&token_get_mutex)
	 *					   return ep->token;	<---------------- awake!
	 *					 )
	 * NOTE:
	 * 		event handler of TU will set ep->token to ECTP_INVALID_TOKEN(-1) if
	 * 		it received TGC with f = 0, else ep->token is set to appropriate
	 * 		token id.
	 */
	ectp_mutex_t token_get_mutex;

	ectp_mutex_t token_ret_mutex;

	/* Token Set
	 * TO must keep the current state of tokens that are assigned to TUs. This
	 * job can be done by using Token Set.
	 *
	 * If TU, this field must be NULL.
	 */ 
	struct ectp_tokenset *tokenset;
};

#define	ECTP_ENDPOINT_LEN	sizeof(struct ectp_endpoint)

/* socket.c ---------------------------------------------------------------- */

/**
 * ectp_create_endpoint() creates ECTP-3 Protocol Endpoint for the given socket
 * structure.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
extern int ectp_create_endpoint( struct ectp_socket *eso, ectp_role_t role );

extern int ectp_destroy_endpoint( struct ectp_endpoint *ep );

/**
 * ectp_ep_alloc() allocates new memory for given ECTP-3 endpoint.
 *
 * RETURN:
 * 		On success ~~ is returned.
 * 		Otherwise NULL is returned.
 */
extern struct ectp_endpoint *ectp_ep_alloc();

/**
 * ectp_ep_free() deallocates memory of the given ECTP-3 endpoint.
 */
extern void ectp_ep_free( struct ectp_endpoint *ep );

static inline int ectp_ep_is_su( struct ectp_endpoint *ep )
{
	return (ep->token > 0 );
}


/**
 * ECTP-3 Connection Structure
 *
 * This contains ECTP-3 Transmission Control Block (TCB) Parameters for two
 * communication channels:
 * 		Forward Multicast Data Channel
 * 		Backward Unicast Data Channel
 *
 * Because there is no boundary beteen data channel and control channel in
 * ECTP-3 specification, there only requires to keep TCB parameters for each
 * channel and there are parameters to manage the control tree, as well as two
 * data channels.
 */

struct ectp_connection {

	/* Backpointer to ECTP-3 endpoint to which this endpoint is beloing. */
	struct ectp_endpoint *ep;

	/* ISO/IEC 14476-3 | ITU-T X.607 */

	/* 8.1 Base Header
	 * 		Version (2bits) :
	 * 		This defines the version of the ECTP-3 protocol.
	 * 		Its current version is encoded as '00'.
	 */	
	__u8	version:2,

	/* 8.1 Base Header
	 * 		CT (Connection Type, 2bits) :
	 * 		01 - simplex multicast connection (for ECTP-1 and ECTP-2);
	 * 		10 - duplex multicast connection (for ECTP-3 and ECTP-4);
	 * 		11 - N-plex multicast connection (for ECTP-5 and ECTP-6);
	 */
			ct:2,

	/* NOTE:
	 * 		TCO, RO, AGN, MSS are given by 'Connection Element' when ECTP-3
	 * 		endpoint receives 'CCR' or 'LJR' packet.
	 * 		Before initial handshaking, they can be set by using
	 * 		msetsockopt() function.
	 */
	
	/* 8.2.1 Connection Element
	 * 		TCO (Tree Configuration Option, 2bits) :
	 * 		00 - Level 1 configuration (No LO, TO <=> TU)
	 * 		01 - Level 2 configuration (TO <=> LO <=> TU or TO <=> TU)
	 * 		10 - General configuration (More than two tree levels)
	 * 		11 - Reserved for future use
	 */
			tco:2,

	/* 8.2.1 Connection Element
	 * 		R (Reserved, 2bits) :
	 */
			r:2;

	/* AGN (ACK Generation Number)
	 * 		Each child generates an ACK packet every ACK Generation Number
	 * 		(AGN) data packets. This is an ACK packet is generated for the AGN
	 * 		data packets. */
	__u8	agn;

	/* MSS (Maximum Segment Size) in [bytes] for each channel. */
	__u16	mss;

	/* 9.3.5 Control Tree Maintenance
	 * 	FDN (Failure Detection Number)
	 * 		A child detects the failure of its parent, if it cannot receive any
	 * 		packets such as HB and RDATA packets from the parent during the
	 * 		time interval of FDN x HGT.
	 * 		A parent detects the failure of a child, if it cannot hear any
	 * 		HBACK packets from the child for the FDN consecutive HB packets.
	 * 		The default FDN is 3.
	 *
	 * 	NOTE:
	 * 		A Child  needs a timer used to detect the failure of its parent.
	 * 		Let's call this a PFW (Parent Failure detection Watchdog).
	 * 		If the PFW expires, a child believes its parent fails.
	 * 		=> Watchdog may not be needed. It can be done by internal event
	 * 		such as ECTP_EVENT_PARENT_FAILURE and its handler function.
	 * 		A Parent needs to keep an expected_cid (Expected Child ID) and
	 * 		polling_cid (Poling Child ID). An expected_cid represents a
	 * 		Child ID that will send a next HBACK packet and polling_cid
	 * 		represents a Child ID that is going to be checked.
	 * 		Usually a parent generates a HB packet with polling_cid when
	 * 		HGT expires. When a parent receives a HBACK, it checks whether
	 * 		the received_cid is same to the expected_cid. If they doesn't
	 * 		match, a parent records this situation into the bit-map as
	 * 		follows:
	 * 				for ( fc = 0; fc < FDN; fc++ ) {
	 * 					if ( cumulative_fail_bitmap[fc][expected_cid] ) {
	 * 						continue;
	 * 					} else {
	 * 						Set cumulative_fail_bitmap[fc][expected_cid];	
	 * 						break;
	 * 					}
	 * 				}
	 * 				if ( fc == FDN-1 )
	 * 					Mark expected_cid as a failure.
	 * 				According to implementation, above bit-map can be
	 * 				represented one-dimensional array as follows:
	 * 					__u32 bitmap[FDN][8];
	 * 					=> Think bitmap[0] as a 256-bit integer.
	 * 		If match, a parent must clear the consecutively recorded
	 * 		information of the expected_cid.
	 */
	__u16	fdn;

	/* Is the parent endpoint connected to ECTP-3 Connection ?
	 * 		0   - Not connected
	 * 		> 0 - Connected
	 */
	__u16	connected;

	/* A.1 Timers
	 * Connection Time [sec]
	 * 		If TO, this field acts as a CCT (Connection Creation Time).
	 * 		TO triggers the CCT timer when sending the CCR packet to the group.
	 * 		The TO will process only the CCC packets that arrive from the
	 * 		prospective TUs before the CCT timer.
	 * 		If TU, this field acts as a CWT (Connection Waiting Time).
	 * 		When a promising late-joining TU starts an ECTP connection, if it
	 * 		cannot receive any CCR message during the CWT time, it will try to
	 * 		join the ECTP connection as a late-joiner by sending an LJR message
	 * 		to TO.
	 *
	 */
	__u16	ctime;


	/* Following fields must be managed seperately at the Multicast / Unicast
	 * Data Channel.
	 */
	struct ectp_channel {
		
		/* RXT [sec] */
		__u16	rxt;

		/* HGT [sec] */
		__u16	hgt;

		/* HB/HBACK is used to maintain control tree. But we need to know Data
		 * channel is still alive. */
		/*__u8	dit; */

		/* ISN (Initial Sequence Number) */
		__u32	isn;

		/* 8.2.2 Acknowledgement Element
		 * 		LSN (Lowest Sequence Number)
		 * 		This represents the sequence number of the lowest numbered DATA
		 * 		pakcet that has not been received yet at the receiver side.
		 * 		It is a Expected value.
		 */
		__u32	lsn;

		/* Recently Received PSN. */
		__u32	rsn;

		/* Recently Sent PSN. */
		__u32	ssn;

	} m_ch, u_ch;

	/* Buffer of the Ack Bitmap */
	struct ectp_ack_buf {
		__u8	next:4,
			   	length:4;
		__u8	vlength;
		__u16	rowpos;
	} ack_buf;

	unsigned long *ack_bitmapbuf;

};

#define ECTP_CONNECTION_LEN	sizeof(struct ectp_connection)

extern int ectp_create_connection( struct ectp_socket *eso,
								   SS *local_addr, SS *group_addr, int cct);

extern int ectp_destroy_connection( struct ectp_socket *eso );

/**
 * ectp_cn_alloc() allocates new memory for given ECTP-3 connection.
 *
 * RETURN:
 * 		On success a new pointer of a ECTP-3 connection is returned.
 * 		Otherwise NULL is returned.
 */
extern struct ectp_connection *ectp_cn_alloc( void );

/**
 * ectp_cn_free() deallocates memory of the given ECTP-3 connection.
 */
extern void ectp_cn_free( struct ectp_connection *cn );


extern void ectp_cn_init( struct ectp_socket *eso,
		struct ectp_connection *cn );


/**
 * ECTP-3 Control Tree
 */

struct ectp_tree {

	/* Backpointer to ECTP-3 endpoint to which this endpoint is beloing. */
	struct ectp_endpoint *ep;

	/* Address of the TO Node */
	ectp_addr_t		to;

	/* Address of the Parent Node (TO or LO)
	 * If ( endpoint->tco == 00 ), parent is TO.
	 * If TO, this field is ignored. */
	ectp_addr_t		parent;
	
	/* NOTE:
	 * 		Tree Level, Child ID, ADN are given by 'Membership Element' when
	 * 		ECTP-3 endpoint receives 'TJR' from its parent (TO or LO).
	 */
	
	/* Tree Level of the Current Node */
	__u8			level;

	/* 8.2.3 Membership Element
	 * 		Child ID (8 bits)
	 * 		This is non-negative integer that specifies the ID number of a
	 * 		child node. In the tree creation phase, the Child ID will be
	 * 		assigned to each of the children by the parent. Child ID is used to
	 * 		determine when to send an ACK packet to its parent. An ACK packet
	 * 		of a child node contaions its Child ID, which can be used for the
	 * 		error recovery by the parent. The value of '0' is used to indicate
	 * 		that this field must be ignored.
	 *
	 * Initial value : ECTP_INVALID_CID (usually 0).
	 *
	 * NOTE:
	 * 		TO does not have its Child ID, therefore, if TO, zero is assigned.
	 */
	ectp_cid_t		cid;

	/* 8.2.3 Membership Element
	 * 		Active Descendant Number (ADN, 16bits)
	 * 		This represents the number of active descendants of the node along
	 * 		the tree. Each TU sets the ADN to '1', since it has no child. The
	 * 		parent LO will aggregate all of the ADN values for its children. In
	 * 		this fashion, TO can obtain the information on how may TUs are
	 * 		participating in the connection.
	 *
	 */
	__u16			adn;

	/* TO/LO must keep the current state of its childrens. This job can be done
	 * by using Child Set.
	 *
	 * IF TU, this field must be NULL.
	 */
	struct ectp_childset *childset;
};

#define	ECTP_TREE_LEN	sizeof(struct ectp_tree)


extern int ectp_create_tree( struct ectp_socket *eso );
extern int ectp_destroy_tree( struct ectp_socket *eso );

extern struct ectp_tree *ectp_tree_alloc( void );
extern void ectp_tree_free( struct ectp_tree *tree );
extern int ectp_tree_init( struct ectp_socket *eso,
						   struct ectp_tree *tree, ectp_role_t role );


/* ------------------------------------------------------------------------- *
 * ACKNOWLEDGEMENT DEFINITION & FUNCTION
 * ------------------------------------------------------------------------- */


#define BITMAP_LEN	sizeof(unsigned long)	// Sizeof 1 Bitmap field 

static inline void bit_set( unsigned long *bitmap, unsigned int bit )
{
	bitmap[ bit / 32 ] |= 0x80000000UL >> ( bit % 32 );
}

static inline void bit_clear( unsigned long *bitmap, unsigned int bit )
{
	bitmap[ bit / 32 ] &= ~( 0x80000000UL >> ( bit % 32 ) );
}

static inline void bit_print( unsigned long *bitmap, int length )
{
	int i = 0;
	int mask = 1, bit;
	int count = 31;

	for( i=0; i<length; i++ ) {
		for( count=31; count>=0; count-- ) {
			bit = (bitmap[i] >> count) & mask;
			printf("%d",bit);
			if( count % 8 == 0 )
				putchar(' ');
		}
		putchar('\n');
	}
}

#ifdef	__cplusplus
}
#endif

#endif	/* _ECTP_STRUCTS_H */
