/* Enhanced Communication Transport Protocol Part-3 Implementation
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * ectp/types.h
 *  -	This header defines various types used by ectp internal functions.
 *
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers <saintpaul1978@mail.knu.ac.kr>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 *
 */

#ifndef _ECTP_TYPES_H
#define _ECTP_TYPES_H

/* -------------------------------------------------------------------------- *
 * HEADER FILES INCLUDED
 * -------------------------------------------------------------------------- */

#include <ectp/wrapper.h>	/* for err_ret() */

#include <sys/types.h>		/* size_t, ssize_t, ... */
#include <sys/socket.h>		/* for socket(), sendto(), ... */
#include <sys/time.h>		/* for gettimeofday() */
#include <time.h>			/* for timeval{} */

#include <netinet/in.h>		/* for sockaddr_in{}, sockaddr_in6{} */

#include <unistd.h>
#include <stdlib.h>			/* for free */
#include <errno.h>			/* for errno global variable */

#include <pthread.h>		/* for posix thread */
#include <semaphore.h>		/* for posix semaphore */

#ifdef	__cplusplus
extern "C" {
#endif



/* -------------------------------------------------------------------------- *
 * Definitions for primitive types
 * -------------------------------------------------------------------------- */

typedef char			__s8;
typedef unsigned char	__u8;

typedef short			__s16;
typedef unsigned short	__u16;

typedef int				__s32;
typedef unsigned int	__u32;

#	if defined(__GNUC__)
typedef long long			__s64;
typedef unsigned long long	__u64;
#	endif



/* -------------------------------------------------------------------------- *
 * Definitions for Synchronization
 * -------------------------------------------------------------------------- */

/* If we use POSIX Semaphore, ... 
 *     int sem_init(sem_t *sem, int pshared, unsigned int value);
 *     int sem_wait(sem_t * sem);
 *     int sem_trywait(sem_t * sem);
 *     int sem_post(sem_t * sem);
 *     int sem_getvalue(sem_t * sem, int * sval);
 *     int sem_destroy(sem_t * sem);
*/

/**
 * ECTP-3 Semaphore structure
 * - We define wrapper structure of the POSIX semaphore.
 *   In this work, we can limit the maximum number of the POSIX semaphore.
 */
typedef struct ectp_sem_struct
{
	sem_t	s;		/* POSIX semaphore */
	/*__u16	max;*/	/* Maximum number of resources => REMOVED! */
	__u32	n;		/* Currently available number of resources  */
} ectp_sem_t;

/* control.c --------------------------------------------------------------- */

/**
 * ectp_seminit() initializes the semaphore object.
 * - On success, 0 is returned. 
 *   Otherwise, -1 is returned.
 */
extern int ectp_sem_init(ectp_sem_t *sem, __u32 val);

/**
 * ectp_sem_destroy() destroys a semaphore object. Once it is destroyed, no
 * threads should be wating on the it.
 * - On success, 0 is returned.
 *   Otherwise, -1 is returned.
 * - If error occurs, check whether errono is EBUSY.
 */
extern int ectp_sem_destroy(ectp_sem_t *sem);

/** upper-bound check function */
/* extern inline int ectp_sem_isfull(ectp_sem_t *sem); */

/**
 * lower-bound check function
 * we can predict if caller will be blocked.
 */ 
static inline int ectp_sem_isempty(ectp_sem_t *sem)
{
	return ( sem->n == 0 );	
}

/**
 * ectp_sem_wait() suspends the calling thread until the semaphore pointed to
 * by sem has non-zero count.
 *
 * RESULT:
 * -	On success 0 is returned.
 *  	Otherwise -1 is returned.
 */
static inline int ectp_sem_wait(ectp_sem_t *sem)
{
	int result = -1;

	do {
		result = sem_wait(&sem->s);
	} while ( result < 0 && errno == EINTR );

	if ( result < 0 ) {
		return -1;
	}

	sem->n--;
	return 0;
}

/**
 * ectp_sem_trywait() is a non-blocking variant of ectp_sem_wait().
 * - This function always return 0.
 */
static inline int ectp_sem_trywait(ectp_sem_t *sem)
{
	sem_trywait(&sem->s);
	sem->n--;
	return 0;
}

/**
 * ectp_sem_post() atomically increases the count of the semaphore pointed to
 * by sem.
 * - On success, 0 is returned.
 *   Otherwise, -1 is returned.
 */
static inline int ectp_sem_post(ectp_sem_t *sem)
{
	if ( sem_post(&sem->s) < 0 )
		err_ret("ectp_sem_post() : sem_post() failed.\n", -1);
	sem->n++;

	return 0;
}

/**
 * Definitions for ECTP-3 mutex type.
 */
typedef ectp_sem_t	ectp_mutex_t;


static inline int ectp_mutex_init(ectp_mutex_t *mutex)
{
	return ectp_sem_init(mutex, 1);
}

/*
 * - This function always return 0.
 */
static inline int ectp_mutex_on(ectp_mutex_t *mutex)
{
	return ectp_sem_wait(mutex);
}

static inline int ectp_mutex_off(ectp_mutex_t *mutex)
{
	return ectp_sem_post(mutex);
}

static inline int ectp_mutex_destroy(ectp_mutex_t *mutex)
{
	return ectp_sem_destroy(mutex);
}



/* ------------------------------------------------------------------------- *
 * Definitions for various data types used ECTP-3 internal functions
 * ------------------------------------------------------------------------- */

/**
 * ECTP-3 status type.
 */
typedef enum 
{
	ECTP_FAILURE		= -1,	/* to support setsockopt() */
#define ECTP_FAILURE	ECTP_FAILURE
	ECTP_SUCCESS		= 0
#define ECTP_SUCCESS	ECTP_SUCCESS
} ectp_status_t;

/**
 * ECTP-3 socket descriptor type.
 */
typedef	__u32 ectp_sockdes_t;

#define	INVALID_SOCKET	-1

/**
 * The role of a node that is taken part in ECTP-3 control tree.
 */
typedef enum
{
	ECTP_ROLE_NONE		= 0,
#define	ECTP_ROLE_NONE	ECTP_ROLE_NONE

	/* 3. Definitions
	 * c)	TO (Top Owner or TC-Owner)
	 * 		TO is single sender of multicast data packets, which can transmit
	 * 		multicst data to the other TS-users, and it manages overall
	 * 		operations of ECTP-3.
	 */
	ECTP_ROLE_TO,
#define ECTP_ROLE_TO	ECTP_ROLE_TO

	/* 3. Definitions
	 * d)	LO (Local Owner)
	 * 		An LO is located on the contree of ECTP-3. One or more LOs could be
	 * 		designated for scalable error recovery and status monitoring in
	 * 		ECTP-3. An LO is anso a TS-user, which can also receive the
	 * 		multicast data from TO. LOs will be configured as a parent of the
	 * 		local groups through the control configuration in ECTP-3.
	 */
	ECTP_ROLE_LO,
#define	ECTP_ROLE_LO	ECTP_ROLE_LO

	/* 3. Definitions
	 * e)	LE (Leaf Entity)
	 * 		An LE is a leaf node on the ECTP-3 control tree. It is a TS-user in
	 * 		the ECTP-3 connection, and it can receive multicast data sent by
	 * 		TO.
	 */
	ECTP_ROLE_LE,
#define	ECTP_ROLE_LE	ECTP_ROLE_LE

	/*	3. Definitions
	 *	a)	SU (Sending User)
	 *		Some of the ECTP-3 TS-users can send unicast data to the TO. A
	 *		sending user (SU) is a TS-user who gets a token from TO. Only the
	 *		SU is allowed to send unicast data to TO. In other words, before
	 *		sending unicast data, each user must request a token to TO.
	 *
	 *	NOTE:
	 *	-	We do not define SU as a ectp_role_t type.
	 *	-	Just use the inline function ectp_ep_is_su().
	 */

} ectp_role_t;

#define	ECTP_ROLE_RA	ECTP_ROLE_LO
#define	ECTP_ROLE_TU	ECTP_ROLE_LE


/*
 * These definitions are for convenience.
 */

#define	SA			struct sockaddr
#define	SS			struct sockaddr_storage
#define	SA_IN		struct sockaddr_in
#define	SA_IN6		struct sockaddr_in6

#define	SA_LEN		sizeof(struct sockaddr)
#define	SS_LEN		sizeof(struct sockaddr_storage)
#define	SA_IN_LEN	sizeof(struct sockaddr_in)
#define	SA_IN6_LEN	sizeof(struct sockaddr_in6)

/**
 * ECTP-3 address type.
 */
typedef union ectp_sockaddr
{
	struct sockaddr_in		sa;
	struct sockaddr_in6		sa6;
} ectp_addr_t;

#define ECTP_ADDR_LEN	sizeof(ectp_addr_t)

static inline socklen_t ectp_addrlen( int domain )
{
	if ( domain == AF_INET )
		return SA_IN_LEN;
	else if ( domain == AF_INET6 )
		return SA_IN6_LEN;

	return 0;
}

static inline SA *ectp_sockaddr( int domain, ectp_addr_t *addr )
{
	if ( domain == AF_INET )
		return (SA *) &addr->sa;
	else if ( domain == AF_INET6 )
		return (SA *) &addr->sa6;

	return NULL;
}

/**
 * Verify if given address is same.
 *
 * NOTE:
 * -	This function only check the domain and corresponding address
 *  	structure. (e.g. sockaddr_in{}->sin_addr, sockaddr_in6{}->sin6_addr)
 *  	That is, the port number is not checked.
 *
 * RETURN:
 * -	This function returns zero, if domain and address are same.
 *  	Otherwize, -1 is returned.
 */
static inline int ectp_addrchk( ectp_addr_t *addr1, ectp_addr_t *addr2 )
{
	int result = -1;

	if ( ((SS *) addr1)->ss_family != ((SS *) addr2)->ss_family )
		return -1;

	switch ( ((SS *) addr1)->ss_family ) {
	case AF_INET:
		result = memcmp(&((SA_IN *) addr1)->sin_addr,
			   		&((SA_IN *) addr2)->sin_addr, sizeof(struct in_addr));
		break;
	case AF_INET6:
		result = memcmp(&((SA_IN6 *) addr1)->sin6_addr,
					&((SA_IN6 *) addr2)->sin6_addr, sizeof(struct in6_addr));
		break;
	default:
		return -1;
	}

	return !result ? 0 : -1;
}



/**
 * 7.5 Tokens
 * In ECTP-3, a token represents the right for a TU to send a unicast data to
 * TO. Before transmitting the data each TU must get a token from the TO.
 *
 * Each token is represented as a 1-byte non-neagtive integer (Token ID). Token
 * ID is ranged between 1 and 255. The Token ID of '0' is reserved for use of
 * TO. At the TO side, the Token ID can be used to identify who is sending the
 * unicast data.
 */
typedef short ectp_token_t;

#define	ECTP_TOKEN_TO			0
#define ECTP_MAX_TOKEN			255
#define	ECTP_INVALID_TOKEN		-1

#define	IS_TOKEN_VALID(t)		( ((t) >= 0) && ((t) <= 255) )

/* Total number of token including Token ID '0'. */
#define	ECTP_NTOKEN				256

/* Bytes per word for the total number of token-words */
#define ECTP_NTOKEN_BPW			32
#define ECTP_NTOKEN_WORDS		( ECTP_NTOKEN / ECTP_NTOKEN_BPW )

/**
 * 8.2.3 Membership Element
 * 		Child ID (8 bits)
 * 		This is non-negative integer that specifies the ID number of a child
 * 		node. In the tree creation phase, the Child ID will be assigned to each
 * 		of the children by the parent. Child ID is used to determine when to
 * 		send an ACK packet to its parent. An ACK packet of a child node
 * 		contaions its Child ID, which can be used for the error recovery by the
 * 		parent. The value of '0' is used to indicate that this field must be
 * 		ignored.
 *
 * Initial value : ECTP_INVALID_CID (usually 0).
 *
 * NOTE:
 * 		TO does not have its Child ID, therefore, if TO, zero is assigned.
 *
 */
typedef short ectp_cid_t;

#define	ECTP_INVALID_CID		0
#define	ECTP_MAX_CID			255

#define	IS_CID_VALID(cid)		( ((cid) >= 0) && ((cid) <= 255) )

#define	ECTP_NCID				256
#define	ECTP_NCID_BPW			32
#define	ECTP_NCID_WORDS			( ECTP_NCID / ECTP_NCID_BPW )


/* ------------------------------------------------------------------------- *
 * ECTP-3 Finite State Machine
 * ------------------------------------------------------------------------- */

typedef	enum
{
	/* Common States */

	ECTP_STATE_CLOSED			= 0,
	ECTP_STATE_ESTABLISHED,
	ECTP_STATE_SHUTDOWN_PENDING,

	ECTP_STATE_TRR_SENT,

	/* TO Only States */

	ECTP_STATE_CR_SENT,
	ECTP_STATE_JR_RCVD,
	ECTP_STATE_TJ_RCVD,	/* App. server must validate the membership. */
	ECTP_STATE_TGR_RCVD,
	
	/* TU Only States */

	ECTP_STATE_CR_WAIT,
	ECTP_STATE_CR_RCVD,
	ECTP_STATE_JR_SENT,
	ECTP_STATE_TJ_SENT,
	ECTP_STATE_TGR_SENT,

} ectp_state_t;

/* For backward compatibility */
#define ECTP_STATE_JOIN_WAIT		ECTP_STATE_CR_WAIT
//#define ECTP_STATE_SHUTDOWN_PENDING	ECTP_STATE_CTR_PENDING
#define ECTP_STATE_CTR_PENDING	ECTP_STATE_SHUTDOWN_PENDING



/* ------------------------------------------------------------------------- *
 * ECTP-3 DEFAULT PROTOCOL PARAMETERS
 * ------------------------------------------------------------------------- */

#define ECTP_PARAM_VER_DFL		0
#define	ECTP_PARAM_CT_DFL		2

#define	ECTP_PARAM_TCO_DFL		0
#define	ECTP_PARAM_AGN_DFL		2
#define	ECTP_PARAM_MSS_DFL		1024	/* [byte] */

#define	ECTP_PARAM_FDN_DFL		3
#define	ECTP_PARAM_CCT_DFL		0		/* [sec] */
#define	ECTP_PARAM_CWT_DFL		0		/* [sec] */
#define	ECTP_PARAM_HGT_DFL		9		/* [sec] */
#define	ECTP_PARAM_RXT_DFL		3		/* [sec] */

#define	ECTP_PARAM_ISN_DFL		1
#define	ECTP_PARAM_LSN_DFL		1

/**
 * ECTP-3 default send/receive buffer size:	64 Kb = 65536 bytes
 *
 * cf.	DEC OSF/1 currently uses a default value of 32768 bytes for TCP send
 * 		and receive buffers.
 *
 */
#define ECTP_PARAM_SNDBUF_DFL	(64 * 1024)
#define ECTP_PARAM_RCVBUF_DFL	(64 * 1024)

///**
// * ECTP-3 option
// */
//#define ECTP_OPT_TCO		1
//#define ECTP_OPT_AGN		2
//#define ECTP_OPT_MSS		3
//#define ECTP_OPT_EVENTS		4


#ifdef	__cplusplus
}
#endif

#endif	/* _ECTP_TYPES_H */

/* vim: set ts=4:sw=4:nonu */
