#include <ectp/types.h>
#include <ectp/wrapper.h>

#include <ectp/structs.h>
#include <ectp/protocol.h>
#include <ectp/event.h>
#include <ectp/ectp.h>

/* -------------------------------------------------------------------------- *
 * Implementations of Synchronization Control Functions
 * -------------------------------------------------------------------------- */

/**
 * ectp_sem_init() initializes the semaphore object.
 * - sem must be malloced before this functions is called.
 * - On success, 0 is returned. 
 *   Otherwise, -1 is returned.
 */
int ectp_sem_init(ectp_sem_t *sem, __u32 val)
{
	if ( !sem )
		err_ret("ectp_sem_init() : null exception on \'sem\'.", -1);

	sem->n = val;

	if ( sem_init(&sem->s, 0, val) < 0 )
		err_ret("ectp_sem_init() : sem_init() failed.\n", -1);

	return 0;
}

/**
 * ectp_sem_destroy() destroys a semaphore object. Once it is destroyed, no
 * threads should be wating on the it.
 * - On success, 0 is returned.
 *   Otherwise, -1 is returned.
 */
int ectp_sem_destroy(ectp_sem_t *sem)
{
	if ( sem_destroy(&sem->s) < 0 )
		err_ret("ectp_sem_destroy() : sem_destroy() failed.\n", -1);

	if ( sem )
		free(sem);
	sem = NULL;

	return 0;
}



/* ------------------------------------------------------------------------- *
 * Token Control Functions
 * ------------------------------------------------------------------------- */

struct ectp_tokenset *ectp_create_tokenset()
{
	struct ectp_tokenset *set = NULL;

	if ( !(set = (struct ectp_tokenset *) calloc(1, ECTP_TOKENSET_LEN)) )
		err_ret( "ectp_create_tokenset() : calloc() failed.", NULL );

	set->n = 0;
	ectp_token_emptyset(set);

	return set;
}

int ectp_destroy_tokenset( struct ectp_tokenset *set )
{
	int token = 0;

	while ( (token = ectp_token_ffs(set)) > 0 ) {
		ectp_token_delset(set, token);
	}

	return 0;
}



/* ------------------------------------------------------------------------- *
 * Child Control Functions
 * ------------------------------------------------------------------------- */

struct ectp_childset *ectp_create_childset()
{
	struct ectp_childset *set = NULL;

	if ( !(set = (struct ectp_childset *) calloc(1, ECTP_CHILDSET_LEN)) )
		err_ret( "ectp_create_childset() : calloc() failed.", NULL );

	set->n = 0;
	ectp_child_emptyset(set);

	return set;
}

int ectp_destroy_childset( struct ectp_childset *set )
{
	int cid = 0;

	while ( (cid = ectp_child_ffs(set)) > 0 ) {
		ectp_child_delset(set, cid);
	}

	return 0;
}




/* ------------------------------------------------------------------------- *
 * EVENT INVOCATION FUNCTIONS
 * ------------------------------------------------------------------------- */

int ectp_ops_start_events( struct ectp_socket *eso )
{
	ectp_role_t role = eso->ep->role;
	struct ectp_handleset *hset = &eso->hset;

	hset->main_loop_is_running = 1;
	if ( pthread_create(&hset->loop, NULL, hset->main_loop, eso ) ) {
		err_goto("ectp_ops_start_events() : "
			"cannot invoke main loop.\n", rb_start_event);
	}

	/* TU start event_handler when join is completely created. */

	/* TO needs additional threads to service concurrently. */
	if ( role == ECTP_ROLE_TO ) {

		hset->event_handler_is_running = 1;
		if ( pthread_create(&hset->event, NULL, hset->event_handler, eso) ) {
			err_goto("ectp_ops_start_events() : "
				"cannot invoke event handler.\n", rb_start_event);
		}

		hset->token_handler_is_running = 1;
		if ( pthread_create(&hset->token, NULL, hset->token_handler, eso) ) {
			err_goto("ectp_ops_start_events() : "
					 "cannot invoke token handler.\n", rb_start_event);
		}

		hset->join_handler_is_running = 1;
		if ( pthread_create(&hset->join, NULL, hset->join_handler, eso) ) {
			err_goto("ectp_ops_start_events() : "
					 "cannot invoke join handler.\n", rb_start_event);
		}
	}

	return 0;

rb_start_event:	/* Roll back point */
	hset->main_loop_is_running		= 0;
	hset->event_handler_is_running	= 0;
	hset->token_handler_is_running	= 0;
	hset->join_handler_is_running	= 0;

	return -1;
}


int ectp_ops_end_events( struct ectp_socket *eso )
{
	ectp_role_t role = eso->ep->role;
	struct ectp_handleset *hset = &eso->hset;

	if ( role == ECTP_ROLE_TO ) {

		hset->token_handler_is_running = 0;
		if ( pthread_cancel(hset->token) != 0 ) {
			err_ret("ectp_ops_end_events() : "
					 "cannot cancel token handler.\n", -1);
		}

		hset->join_handler_is_running = 0;
		if ( pthread_cancel(hset->join) != 0 ) {
			err_ret("ectp_ops_end_events() : "
					 "cannot cancel join handler.\n", -1);
		}
	}

	hset->event_handler_is_running = 0;
	if ( pthread_cancel(hset->event) != 0 ) {
		err_ret("ectp_ops_end_events() : "
			"cannot cancel event handler.\n", -1);
	}

	hset->main_loop_is_running = 0;
	if ( pthread_cancel(hset->loop) != 0 ) {
		err_ret("ectp_ops_end_events() : "
			"cannot cancel main loop.\n", -1);
	}

	return 0;
}



/* ------------------------------------------------------------------------- *
 * STATE MACHINE CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

/**
 */
int ectp_ops_connection_creation( struct ectp_socket *eso, int time )
{
	/* Not completely implemented. */

	/*
	int size = sizeof(struct ectp_packet_ccr);

	char *buff = (char *)malloc(size);
	memset( buff, 0, size );
	struct ectp_packet_ccr *packet;
	packet = (struct ectp_packet_ccr*) buff;

	ectp_make_packet( (void *)packet, ECTP_PKT_CR );

	sendto( esd, buff, size, 0, (struct sockaddr*)&eso->eso_groupaddr.g_addr,
			sizeof(struct sockaddr) );
	*/

	/* Actually, when CCT expires, state can be set to ESTABLISHED. */
	eso->ep->state = ECTP_STATE_ESTABLISHED;

	return 0;
}

/**
 * Start Late Join handshake.
 *
 * 		1	= Accept
 * 		0	= Deny
 * 		-1	= Error
 */
int ectp_ops_start_latejoin( struct ectp_socket *eso, int time )
{
	struct ectp_endpoint *ep = eso->ep;
	struct ectp_connection *cn = ep->cn;
	struct ectp_tree *tree = ep->tree;
	ectp_pkt_jr	*jr = NULL;
	ectp_pkt_jc	*jc = NULL;
	ectp_pkt_tj	*tj = NULL;
	ectp_pkt_tc	*tc = NULL;
	struct ectp_skbuff *skb = NULL;
	struct ectp_timeritem *rxt = NULL;
	struct ectp_eventitem *e = NULL;

	socklen_t addrlen;

	/* Waiting Connection Creation! (STATE: CR-WAIT) */
	puts("ECTP-3, Waiting Connection Creation Request! (STATE: CR-WAIT)");
	ep->state = ECTP_STATE_CR_WAIT;

	/* Making an new Alarm Timer for CWT
	 * Skip CR-WAIT manipulation. */

	/* make JR */
	if ( !(jr = ectp_create_jr(eso)) )
		err_ret("ectp_ops_start_latejoin() : cannot create JR.", -1);

	/* Makes send buff item. */
	addrlen = ectp_addrlen(eso->domain);
	skb = ectp_skb_alloc(eso, eso->u_sd, &eso->local_addr, addrlen);
	if ( !skb )
		err_ret("ectp_ops_start_latejoin() : ectp_skb_alloc() failed.", -1);

	skb->pktlen	= ECTP_PKT_JR_LEN;
	skb->pkt	= jr;
	skb->psn	= jr->psn;

	ectp_skblist_add( &eso->u_sndbuf, skb );

	/* make RXT timer */
	rxt = ectp_create_timer(ECTP_TIMER_RXT, cn->u_ch.rxt, skb);
	ectp_timerlist_add(&eso->timer_list, rxt);

	ectp_skb_transmit(skb);

	puts("ECTP-3, Waiting Late Join Confirm! (STATE: JR-SENT)");
	ep->state = ECTP_STATE_JR_SENT;

	while ( 1 ) {
		e = ectp_event_deq( &eso->eventq );

		/* block: waiting for JC ! */

		if ( e->type == ECTP_CON_LATE_JOIN ) {
			ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));
			break;
		}

		else if ( e->type == ECTP_EXP_LATE_JOIN ) {
			/* RXT for JR is already removed by ectp_timer_dispatch(). */
			ectp_destroy_event( e );
			ep->state = ECTP_STATE_CLOSED;
			err_ret("ECTP-3, Late Join Failed! (STATE: CLOSED)\n"
					"ectp_ops_start_latejoin() : "
					"Exceed maximum retransmission for JR.", -1);
		}

		else if ( e->type == ECTP_DNY_LATE_JOIN ) {
			ectp_destroy_event( e );
			ep->state = ECTP_STATE_CLOSED;
			ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));
			err_ret("ECTP-3, Late Join Failed! (STATE: CLOSED)\n"
					"ectp_ops_start_latejoin() : "
					"Request for late join is denied.", 0);
		}

		else {
			/* Can't enter here. */
		}
	} /* while ( 1 ) */

	/* If receives JC with f=1 ... */

	jc = (ectp_pkt_jc *) e->packet;

	/* Updates connection info. */
	cn->tco	= jc->connection.tco;
	cn->agn	= jc->connection.agn;
	cn->mss	= jc->connection.mss;

	/* Updates tree info. */
	if ( jc->connection.next == ECTP_EE_NOELEMENT ) {
		/* No additional optional field. Therefore, TO is the parent node. */
		memcpy(&tree->parent, &eso->local_addr, ECTP_ADDR_LEN);
	} else if ( jc->connection.next == ECTP_EE_ADDRESS ) {
		/* Sets specified parent into tree info. */
	} else {
		err_ret("ectp_ops_start_latejoin() : JC next parameter error", -1);
	}

	/* JR terminate. */
	if ( e ) {
		ectp_destroy_event( e );
		e = NULL;
	}


	/* Try to 'tree-join' to 'tree->parent' */

	if ( !(tj = ectp_create_tj(eso)) )
		err_ret("ectp_ops_start_latejoin() : cannot create TJ.", -1);

	/* Makes send buff item. */
	addrlen = ectp_addrlen(((SA *) &tree->parent)->sa_family);
	if ( !(skb = ectp_skb_alloc(eso, eso->u_sd, &tree->parent, addrlen)) )
		err_ret("ectp_ops_start_latejoin() : ectp_skb_alloc() failed.", -1);

	skb->pktlen	= ECTP_PKT_TJ_LEN;
	skb->pkt	= tj;
	skb->psn	= tj->psn;

	ectp_skblist_add( &eso->u_sndbuf, skb );

	/* Makes RXT timer. */

	rxt = ectp_create_timer(ECTP_TIMER_RXT, cn->u_ch.rxt, skb);
	ectp_timerlist_add(&eso->timer_list, rxt);

	ectp_skb_transmit(skb);

	puts("ECTP-3, Waiting Tree Join Confirm! (STATE: TJ-SENT)");
	ep->state = ECTP_STATE_TJ_SENT;

	while ( 1 ) {
		e = ectp_event_deq( &eso->eventq );

		/* block: waiting for TC ! */

		if ( e->type == ECTP_CON_TREE_JOIN ) {
			ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));
			break;
		}

		else if ( e->type == ECTP_EXP_TREE_JOIN ) {
			/* RXT for TJ is already removed by ectp_timer_dispatch(). */
			ectp_destroy_event( e );
			ep->state = ECTP_STATE_CLOSED;
			err_ret("ECTP-3, Tree Join Failed! (STATE: CLOSED)\n"
					"ectp_ops_start_latejoin() : "
					"Exceed maximum retransmission for TJ.", -1);
		}

		else if ( e->type == ECTP_DNY_TREE_JOIN ) {
			ectp_destroy_event( e );
			ep->state = ECTP_STATE_CLOSED;
			ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));
			err_ret("ECTP-3, Tree Join Failed! (STATE: CLOSED)\n"
					"ectp_ops_start_latejoin() : "
					"Request for tree join is denied.", 0);
		}

		else {
			/* Cannot enter here. */
		}
	} /* while ( 1 ) */

	/* If receives TC with f=1 ... */

	tc = (ectp_pkt_tc *) e->packet;

	/* Updates connection info. */

	/* Updates tree info. */
	tree->level	= tc->membership.tree_level;
	tree->cid	= tc->membership.child_id;
	tree->adn	= tc->membership.adn;

	/* TJ terminate. */
	if ( e ) {
		ectp_destroy_event( e );
		e = NULL;
	}


	puts("ECTP-3, Late Join Completed! (STATE: ESTABLISHED)");
	ep->state = ECTP_STATE_ESTABLISHED;

	/* Invokes event handler. */
	eso->hset.event_handler_is_running = 1;
	if ( pthread_create(&eso->hset.event, NULL, eso->hset.event_handler, eso) )
		err_ret("ectp_ops_start_latejoin() : cannot invoke event handler.", -1);

	return 1;
}

/**
 * When the parent(TO/LO) do 'Toublemaker Rejection', this function is called
 * with the specified child address 'to' by the parent.
 *
 * On the other hand, when the local endpoint starts 'User Leave Process', this
 * function is called by upper layer application user. In this case to must be
 * set to NULL.
 */
int ectp_ops_user_leave( struct ectp_socket *eso,
		ectp_addr_t *to, socklen_t tolen )
{
	struct ectp_tree *tree = eso->ep->tree;
	ectp_pkt_lr *lr = NULL;

	/* 'Troublemake Rejection', LR with f=0 is sent to 'to' */
	if ( to ) {
		printf( "Troublemake Rejection - sending IP : \n" );
		if ( !(lr = ectp_create_lr(eso, 0))	)
			err_ret("ectp_ops_user_leave() : cannot create LR.", -1);
		ectp_pkt_transmit(eso->u_sd, lr, ECTP_PKT_LR_LEN, to, tolen);
	}

	/* 'User Leave Process', LR with f=1 is sent to 'to' */
	else {
		if ( !(lr = ectp_create_lr(eso, 1))	)
			err_ret("ectp_ops_user_leave() : cannot create LR.", -1);
		ectp_pkt_transmit(eso->u_sd, lr, ECTP_PKT_LR_LEN,
				&eso->ep->tree->parent,
				ectp_addrlen(((SA *) &tree->parent)->sa_family));

	}

	return 0;
}

int ectp_ops_request_userleave( struct ectp_socket *eso, 
		ectp_addr_t *from, socklen_t fromlen )
{
	struct ectp_childset *childset = eso->ep->tree->childset;
	int result;
	int num;

	for ( num = 0; num < ECTP_NCID; num++ )
	{
		if ( childset->child_addrs[num] != NULL )
			result = memcmp( from, childset->child_addrs[num], fromlen );
		if ( result == 0 )
			break;
	}

	if ( num != ECTP_NCID ) {
		ectp_child_delset( childset, num );
	}
	else
		err_ret("ectp_ops_request_userleave() : cannot find child addr", -1 );

	return 0;
}

int ectp_ops_conn_terminate( struct ectp_socket *eso, int f )
{
	ectp_pkt_ct *ct = NULL;

	if ( !(ct = ectp_create_ct(eso, f)) )
		err_ret("ectp_ops_conn_terminate() : cannot create CT.", -1);

	return ectp_pkt_transmit(eso->m_sd, ct, ECTP_PKT_CT_LEN,
			&eso->group_addr, ectp_addrlen(eso->domain) );
}

int ectp_ops_request_tokenget( struct ectp_socket *eso )
{
	ectp_pkt_tgr *tgr = NULL;
	struct ectp_skbuff *skb = NULL;
	struct ectp_timeritem *rxt = NULL;

	if ( !(tgr = ectp_create_tgr(eso)) )
		err_ret("ectp_ops_request_tokenget() : cannot create TGR.", -1);

	if ( !(skb = ectp_skb_alloc(eso, eso->u_sd,
					&eso->local_addr, ectp_addrlen(eso->domain))) )
		err_ret("ectp_ops_request_tokenget() : ectp_skb_alloc() failed.", -1);

	skb->pktlen	= ECTP_PKT_TGR_LEN;
	skb->pkt	= tgr;
	skb->psn	= tgr->psn;

	ectp_skblist_add( &eso->u_sndbuf, skb );

	/* make RXT timer */
	rxt = ectp_create_timer(ECTP_TIMER_RXT, eso->ep->cn->u_ch.rxt, skb);
	ectp_timerlist_add(&eso->timer_list, rxt);

	/* transmit TGR */
	ectp_skb_transmit(skb);

	ectp_mutex_on(&eso->ep->token_get_mutex);

	/* Waiting for TGC ... */

	ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));

	return eso->ep->token;

	/* If TGC with F = 0, token may be ECTP_INVALID_TOKEN(-1). */
}

int ectp_ops_request_tokenreturn( struct ectp_socket *eso )
{
	ectp_pkt_trr *trr = NULL;
	struct ectp_skbuff *skb = NULL;
	struct ectp_timeritem *rxt = NULL;

	if ( !(trr = ectp_create_trr(eso, eso->ep->token)) )
		err_ret("ectp_ops_request_tokenreturn() : cannot create TRR.", -1);

	if ( !(skb = ectp_skb_alloc(eso, eso->u_sd,
					&eso->local_addr, ectp_addrlen(eso->domain))) )
		err_ret("ectp_ops_request_tokenrturn() : ectp_skb_alloc() failed.", -1);

	skb->pktlen		= ECTP_PKT_TRR_LEN;
	skb->pkt		= trr;

	ectp_skblist_add( &eso->u_sndbuf, skb );

	/* make RXT timer */
	rxt = ectp_create_timer(ECTP_TIMER_RXT, eso->ep->cn->u_ch.rxt, skb);
	ectp_timerlist_add(&eso->timer_list, rxt);

	/* transmit TRR */
	ectp_skb_transmit(skb);

	ectp_mutex_on(&eso->ep->token_ret_mutex);

	/* Waiting for TRC ... */

	ectp_destroy_timer(ectp_timerlist_del(&eso->timer_list, rxt));

	return 0;

	/* Token is returned successfully. */
}

int ectp_ops_ack( struct ectp_socket *eso )
{
	ectp_pkt_ack *ack = NULL;
	ssize_t	len = eso->ep->cn->ack_buf.length;

	if ( !(ack = ectp_create_ack(eso)) ) {
		err_ret("ectp_ops_ack() : ectp_create_ack() failed.", -1);
	}

	memcpy(ack + 1, eso->ep->cn->ack_bitmapbuf, len);

	ectp_pkt_transmit( eso->u_sd, ack, ECTP_PKT_ACK_LEN + len,
						&eso->local_addr, ectp_addrlen(eso->domain));

	return 0;
}

int ectp_ops_accept_latejoin( struct ectp_socket *eso, ectp_addr_t *from )
{
	/* send LJC with F = 1 */
	ectp_pkt_jc *jc = NULL;
	
	if ( !(jc = ectp_create_jc(eso, 1, 0)) )
		err_ret("ectp_ops_accept_latejoin() : ectp_create_jc() failed.", -1);

	ectp_pkt_transmit(
			eso->u_sd, jc, ECTP_PKT_JC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}

int ectp_ops_deny_latejoin( struct ectp_socket *eso, ectp_addr_t *from )
{
	/* send LJC with F = 0 */
	ectp_pkt_jc *jc = NULL;
	
	if ( !(jc = ectp_create_jc(eso, 0, 0)) )
		err_ret("ectp_ops_deny_latejoin() : ectp_create_jc() failed.", -1);

	ectp_pkt_transmit(
			eso->u_sd, jc, ECTP_PKT_JC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}

int ectp_ops_accept_treejoin( struct ectp_socket *eso, ectp_addr_t *from )
{
	/* send TC with F = 1 */
	ectp_cid_t cid;
	ectp_pkt_tc *tc = NULL;
	
	if ( (cid = ectp_child_ffz(eso->ep->tree->childset)) < 0 )
		err_ret("ectp_ops_accept_treejoin() : child full!", -1);

	if ( !(tc = ectp_create_tc(eso, 1, 0)) )
		err_ret("ectp_ops_accept_treejoin() : ectp_create_tc() failed.", -1);

	tc->membership.child_id = cid;

	ectp_child_addset( eso->ep->tree->childset, cid, from );

	ectp_pkt_transmit(
			eso->u_sd, tc, ECTP_PKT_TC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}

int ectp_ops_deny_treejoin( struct ectp_socket *eso, ectp_addr_t *from )
{
	/* send TC with F = 0 */
	ectp_pkt_tc *tc = NULL;
	
	if ( !(tc = ectp_create_tc(eso, 0, 0)) )
		err_ret("ectp_ops_deny_treejoin() : ectp_create_tc() failed.", -1);

	ectp_pkt_transmit(
			eso->u_sd, tc, ECTP_PKT_TC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}


/*
 * RETURN:
 * 		-1	-	error
 * 		0	-	token full
 * 		positive token id	-	OK
 */
int ectp_ops_accept_tokenget( struct ectp_socket *eso, ectp_addr_t *from )
{

	/* send TGC with F = 1 */
	ectp_pkt_tgc *tgc = NULL;
	ectp_token_t token;

	/* token is assigned by TO. */
	if ( (token = ectp_token_ffz(eso->ep->tokenset)) < 0 )
		err_ret("ectp_ops_accept_tokenget() : token full!", 0);

	if ( !(tgc = ectp_create_tgc(eso, 1)) )
		err_ret("ectp_ops_accept_tokenget() : ectp_create_tgc() failed.", -1);
	tgc->token_id = (__u8) token;

	ectp_token_addset(eso->ep->tokenset, tgc->token_id, from);

	ectp_pkt_transmit(
			eso->u_sd, tgc, ECTP_PKT_TGC_LEN, from, ectp_addrlen(eso->domain));

	return token;
}

int ectp_ops_deny_tokenget( struct ectp_socket *eso, ectp_addr_t *from )
{
	/* send TGC with F = 0 */
	ectp_pkt_tgc *tgc = NULL;

	if ( !(tgc = ectp_create_tgc(eso, 0)) )
		err_ret("ectp_ops_deny_tokenget() : ectp_create_tgc() failed.", -1);

	ectp_pkt_transmit(
			eso->u_sd, tgc, ECTP_PKT_TGC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}

int ectp_ops_accept_tokenreturn( struct ectp_socket *eso, 
								ectp_token_t token, ectp_addr_t *from )
{
	/* send TRC with F = 1 */
	ectp_pkt_trc *trc = NULL;

	if ( !(trc = ectp_create_trc(eso, 1)) )
		err_ret("ectp_ops_accept_tokenreturn() : ectp_create_trc() failed.", -1);

	ectp_token_delset(eso->ep->tokenset, token );

	ectp_pkt_transmit(
			eso->u_sd, trc, ECTP_PKT_TRC_LEN, from, ectp_addrlen(eso->domain));

	return 0;
}




/* ------------------------------------------------------------------------- *
 * INPUT/OUTPUT CONTROL FUNCTIONS
 * ------------------------------------------------------------------------- */

ectp_pkt_ct *ectp_create_ct(struct ectp_socket *eso, int f)
{
	ectp_pkt_ct *ct = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	ct = (ectp_pkt_ct *) ectp_pkt_alloc(ECTP_PKT_CT, 0, 0);

	if ( !ct )
		err_ret("ectp_create_ct() : ectp_pkt_alloc() failed.", NULL);

	ct->next			= ECTP_EE_NOELEMENT;
	ct->version			= cn->version;
	ct->ct				= cn->ct;
	ct->packet_type		= ECTP_PKT_CT;
	ct->conn_id			= 0;
	/* f: '0' for the normal termination, '1' for an abnormal termination */
	ct->f				= f;

	/* psn, payload_lengh, token_id are set to zero and ignored. */

	return ct;
}

ectp_pkt_lr *ectp_create_lr(struct ectp_socket *eso, int f)
{
	ectp_pkt_lr *lr = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	lr = (ectp_pkt_lr *) ectp_pkt_alloc(ECTP_PKT_LR, 0, 0);
	if ( !lr )
		err_ret("ectp_create_lr() : ectp_pkt_alloc() failed.", NULL);

	lr->next			= ECTP_EE_NOELEMENT;
	lr->version			= cn->version;
	lr->ct				= cn->ct;
	lr->packet_type		= ECTP_PKT_LR;
	lr->conn_id			= 0;
	/* f: '1 for the user-invoked leave, or '0' for the troublemaker ejection */
	lr->f				= f;

	/* psn, payload_length, token_id are set to zero and ignored. */

	return lr;
}

ectp_pkt_dt	*ectp_create_dt( struct ectp_socket *eso, 
								int payloadlen, int flags )
{
	ectp_pkt_dt	*dt = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	dt = (ectp_pkt_dt *) ectp_pkt_alloc(ECTP_PKT_DT, payloadlen, 0);

	if ( !dt )
		err_ret("ectp_create_dt() : ectp_pkt_alloc() failed.", NULL);

	dt->next			= ECTP_EE_NOELEMENT;
	dt->version			= cn->version;
	dt->ct				= cn->ct;
	dt->packet_type		= ECTP_PKT_DT;
	/* checksum */
	dt->conn_id			= 0;
	/* psn */
	dt->payload_length	= payloadlen;
	/* f */
	/* reserved */
	dt->token_id		= eso->ep->token;

	/* f is set to zero and ignored. */

	switch ( flags ) {

	case ECTP_UCAST:
		dt->psn			= ++cn->u_ch.ssn;
		break;
	case ECTP_MCAST:
		dt->psn			= ++cn->m_ch.ssn;
		break;
	}

	return dt;
}

ectp_pkt_ack *ectp_create_ack( struct ectp_socket *eso )
{
	ectp_pkt_ack *ack = NULL;
	struct ectp_connection *cn = eso->ep->cn;
	struct ectp_tree *tree = eso->ep->tree;
	__u16  length = cn->ack_buf.length;

	ack = (ectp_pkt_ack *) ectp_pkt_alloc(ECTP_PKT_ACK, length, 0);
	length += ECTP_EE_MEMBERSHIP_LEN;

	if ( !ack )
		err_ret("ectp_create_ack() : ectp_pkt_alloc() failed.", NULL);

	ack->basehdr.next			= ECTP_EE_MEMBERSHIP;
	ack->basehdr.version		= cn->version;
	ack->basehdr.ct				= cn->ct;
	ack->basehdr.packet_type	= ECTP_PKT_ACK;
	ack->basehdr.conn_id		= 0;
	ack->basehdr.psn			= cn->m_ch.lsn;
	ack->basehdr.payload_length	= length;
	ack->basehdr.token_id		= eso->ep->token;

	ack->membership.next		= ECTP_EE_ACKNOLEDGEMENT;
	ack->membership.tree_level	= tree->level;
	ack->membership.child_id	= tree->cid;
	ack->membership.adn			= tree->adn;

	ack->acknowledgement.next	= ECTP_EE_NOELEMENT;
	ack->acknowledgement.bitmap_length = cn->ack_buf.length;
	ack->acknowledgement.valid_bitmap = cn->ack_buf.vlength;

	return ack;
}

ectp_pkt_jr *ectp_create_jr( struct ectp_socket *eso )
{
	ectp_pkt_jr *jr = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	jr = (ectp_pkt_jr *) ectp_pkt_alloc(ECTP_PKT_JR, 0, 0);

	if ( !jr )
		err_ret("ectp_create_jr() : ectp_pkt_alloc() failed.", NULL);

	jr->next			= ECTP_EE_NOELEMENT;
	jr->version			= cn->version;
	jr->ct				= cn->ct;
	jr->packet_type		= ECTP_PKT_JR;
	jr->conn_id			= 0;

	/* psn, payload_length, f, token_id are set to zero and ignored. */

	return jr;
}

ectp_pkt_jc *ectp_create_jc( struct ectp_socket *eso, int f, int option )
{
	ectp_pkt_jc *jc = NULL;
	struct ectp_connection *cn = eso->ep->cn;
	__u16  length;


	if ( !option ) {
		jc = (ectp_pkt_jc *) ectp_pkt_alloc(ECTP_PKT_JC, 0, 0);
		length = ECTP_EE_CONNECTION_LEN;
	} else {
		jc = (ectp_pkt_jc *) ectp_pkt_alloc(
				ECTP_PKT_JC, option, eso->domain);
		/* length = ECTP_EE_CONNECTION_LEN [+ address element size] */
		length = ECTP_EE_CONNECTION_LEN + option;
	}

	if ( !jc )
		err_ret("ectp_create_jc() : ectp_pkt_alloc() failed.", NULL);

	jc->basehdr.next			= ECTP_EE_CONNECTION;
	jc->basehdr.version			= cn->version;
	jc->basehdr.ct				= cn->ct;
	jc->basehdr.packet_type		= ECTP_PKT_JC;
	jc->basehdr.conn_id			= 0;
	jc->basehdr.payload_length	= length;
	jc->basehdr.f				= f;

	jc->connection.next			= ECTP_EE_NOELEMENT;
	jc->connection.tco			= cn->tco;
	jc->connection.agn			= cn->agn;
	jc->connection.mss			= cn->mss;

	return jc;
}

ectp_pkt_tj *ectp_create_tj(struct ectp_socket *eso)
{
	ectp_pkt_tj *tj = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	tj = (ectp_pkt_tj *) ectp_pkt_alloc(ECTP_PKT_TJ, 0, 0);
	if ( !tj )
		err_ret("ectp_create_tj() : ectp_pkt_alloc() failed.", NULL);

	tj->next			= ECTP_EE_NOELEMENT;
	tj->version			= cn->version;
	tj->ct				= cn->ct;
	tj->packet_type		= ECTP_PKT_TJ;
	tj->conn_id			= 0;

	/* psn, payload_length, f, token_id are set to zero and ignored. */

	return tj;
}

/**
 * F:	'1' if the TJR request is accepted, '0' otherwise.
 */
ectp_pkt_tc *ectp_create_tc(struct ectp_socket *eso, int f, int option)
{
	ectp_pkt_tc *tc = NULL;
	struct ectp_connection *cn = eso->ep->cn;
	struct ectp_tree *tree = eso->ep->tree;
	__u16  length;


	if ( !option ) {
		tc = (ectp_pkt_tc *) ectp_pkt_alloc(ECTP_PKT_TC, 0, 0);
		length = ECTP_EE_MEMBERSHIP_LEN;
	} else {
		tc = (ectp_pkt_tc *) ectp_pkt_alloc(
				ECTP_PKT_TC, 1, eso->domain);
		/* length = ECTP_EE_MEMBERSHIP_LEN [+ address element size] */
	}

	if ( !tc )
		err_ret("ectp_create_tc() : ectp_pkt_alloc() failed.", NULL);

	tc->basehdr.next			= ECTP_EE_MEMBERSHIP;
	tc->basehdr.version			= cn->version;
	tc->basehdr.ct				= cn->ct;
	tc->basehdr.packet_type		= ECTP_PKT_TC;
	tc->basehdr.conn_id			= 0;
	tc->basehdr.payload_length	= length;
	tc->basehdr.f				= f;

	tc->membership.next			= ECTP_EE_NOELEMENT;
	tc->membership.tree_level	= tree->level;

	/* child_id	is assigned by the parent node. */

	return tc;
}

ectp_pkt_tgr *ectp_create_tgr(struct ectp_socket *eso)
{
	ectp_pkt_tgr *tgr = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	tgr = (ectp_pkt_tgr *) ectp_pkt_alloc(ECTP_PKT_TGR, 0, 0);

	if ( !tgr )
		err_ret("ectp_create_tgr() : ectp_pkt_alloc() failed.", NULL);

	tgr->next			= ECTP_EE_NOELEMENT;
	tgr->version		= cn->version;
	tgr->ct				= cn->ct;
	tgr->packet_type	= ECTP_PKT_TGR;
	tgr->conn_id		= 0;
	/* PSN: the initial PSN of the backward data channel; */
	tgr->psn			= cn->u_ch.isn;

	/* payload_length, f, token_id are set to zero and ignored. */

	return tgr;
}

ectp_pkt_tgc *ectp_create_tgc( struct ectp_socket *eso, int f )
{
	ectp_pkt_tgc *tgc = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	tgc = (ectp_pkt_tgc *) ectp_pkt_alloc(ECTP_PKT_TGC, 0, 0);

	if ( !tgc )
		err_ret("ectp_create_tgc() : ectp_pkt_alloc() failed.", NULL);

	tgc->next			= ECTP_EE_NOELEMENT;
	tgc->version		= cn->version;
	tgc->ct				= cn->ct;
	tgc->packet_type	= ECTP_PKT_TGC;
	tgc->conn_id		= 0;
	tgc->f				= f;

	/* token_id is allocated by TO */

	return tgc;
}

ectp_pkt_trr *ectp_create_trr( struct ectp_socket *eso, int token_id )
{
	ectp_pkt_trr *trr = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	trr = (ectp_pkt_trr *) ectp_pkt_alloc(ECTP_PKT_TRR, 0, 0);

	if ( !trr )
		err_ret("ectp_create_trr() : ectp_pkt_alloc() failed.", NULL);

	trr->next			= ECTP_EE_NOELEMENT;
	trr->version		= cn->version;
	trr->ct				= cn->ct;
	trr->packet_type	= ECTP_PKT_TRR;
	trr->conn_id		= 0;
	trr->token_id		= token_id;

	/* token_id is the Token ID for TU to be returned. */

	return trr;
}

ectp_pkt_trc *ectp_create_trc( struct ectp_socket *eso, int f )
{
	ectp_pkt_trc *trc = NULL;
	struct ectp_connection *cn = eso->ep->cn;

	trc = (ectp_pkt_trc *) ectp_pkt_alloc(ECTP_PKT_TRC, 0, 0);

	if ( !trc )
		err_ret("ectp_create_trc() : ectp_pkt_alloc() failed.", NULL);

	trc->next			= ECTP_EE_NOELEMENT;
	trc->version		= cn->version;
	trc->ct				= cn->ct;
	trc->packet_type	= ECTP_PKT_TRC;
	trc->conn_id		= 0;
	trc->f				= f;

	/* token is the Token ID of TU to be returned. */

	return trc;
}

/**
 * @param optlen	indicates the length in bytes of optional fields/user data.
 * @param domain	indicates whether the address family is IPv4 or IPv6.
 * 					If domain is zero, it will be ignored.
 */
void *ectp_pkt_alloc( ectp_packet_t type, int optlen, int domain )
{
	void *pkt = NULL;
	size_t pktlen = 0;

	switch ( type ) {
	case ECTP_PKT_CC:
	case ECTP_PKT_TJ:
	case ECTP_PKT_ND:
	case ECTP_PKT_JR:
	case ECTP_PKT_LR:
	case ECTP_PKT_CT:
	case ECTP_PKT_TGR:
	case ECTP_PKT_TGC:
	case ECTP_PKT_TRR:
	case ECTP_PKT_TRC:
		pktlen = ECTP_BASEHDR_LEN;
		break;
	case ECTP_PKT_ACK:
		pktlen = ECTP_BASEHDR_LEN + ECTP_EE_MEMBERSHIP_LEN + optlen;
		break;
	case ECTP_PKT_CR:
		pktlen = ECTP_BASEHDR_LEN + ECTP_EE_CONNECTION_LEN;
		break;
	case ECTP_PKT_TC:
		pktlen = ECTP_BASEHDR_LEN + ECTP_EE_MEMBERSHIP_LEN;
		break;
	case ECTP_PKT_DT:
	case ECTP_PKT_RD:
		pktlen = ECTP_BASEHDR_LEN + optlen;
		break;
	case ECTP_PKT_HB:
	case ECTP_PKT_HBACK:
		break;
	case ECTP_PKT_JC:
		pktlen = ECTP_PKT_JC_LEN + optlen;
		break;
	default:
		break;
	}

	pkt = malloc(pktlen);
	if ( !pkt )
		err_ret("ectp_pkt_alloc() : calloc() failed.", NULL);

	memset( pkt, 0, pktlen );

	return pkt;
}

ssize_t ectp_ops_recvmsg(struct ectp_socket *eso, void *msg, size_t len,
		int flags, ectp_addr_t *from, socklen_t *fromlen)
{
	struct ectp_skbuff *skb = NULL;
	ectp_pkt_dt *data = NULL;
	__u8 *payload = NULL;
	__u16 payloadlen;
	ssize_t nbytes = 0;

	if ( flags == ECTP_MCAST ) {
		skb = ectp_skblist_del(&eso->m_rcvbuf, NULL);
	} else if ( flags == ECTP_UCAST ) {
		skb = ectp_skblist_del(&eso->u_rcvbuf, NULL);
	} else {
		/* Can't enter here. See mrecvmsg() at user.c. */
	}

	/* bocking ... */


	if ( !skb )
		err_ret("ectp_ops_recvmsg() : ectp_skblist_del() failed.", -1);

	data = (ectp_pkt_dt *) skb->pkt;
	payloadlen = data->payload_length;
	payload = (__u8 *) data;
	payload += ECTP_BASEHDR_LEN;

	if ( payloadlen <= len ) {	/* payloadlen <= len */
		nbytes = payloadlen;
		memcpy(msg, payload, payloadlen);
	} else {
		nbytes = len;
		printf( "payload : %d, len : %d\n", payloadlen, len );
		memcpy(msg, payload, len);
		err_msg("ectp_ops_recvmsg() : Warning, given length is not sufficient"
				"                     to contain the hole message.\n");
	}

	if ( from != NULL && fromlen != NULL ) {
		memcpy(from, &skb->addr, *fromlen);
		*fromlen = skb->addrlen;
	}

	ectp_skb_free(skb);

	return nbytes;
}

ssize_t ectp_ops_sendmsg(struct ectp_socket *eso, const void *msg, size_t len,
		int flags, ectp_addr_t *to, socklen_t tolen)
{
	ectp_pkt_dt *dt = NULL;
	struct ectp_skbuff *skb = NULL;
	socklen_t addrlen;

	if ( !(dt = ectp_create_dt(eso, len, flags)) ) {
		err_ret("ectp_ops_sendmsg() : ectp_create_dt() failed.", -1);
	}

	memcpy(dt + 1, msg, len);
	addrlen = ectp_addrlen(eso->domain);

	switch (flags) {

	case ECTP_UCAST:
		skb = ectp_skb_alloc(eso, eso->u_sd, &eso->local_addr, addrlen);
		if ( !skb )
			err_ret("ectp_ops_sendmsg() : ectp_skb_alloc() failed.", -1);

		skb->pktlen	= ECTP_PKT_DT_LEN + len;
		skb->pkt	= dt;
		skb->psn	= dt->psn;

		break;

	case ECTP_MCAST:
		skb = ectp_skb_alloc(eso, eso->m_sd, &eso->group_addr, addrlen);
		if ( !skb )
			err_ret("ectp_ops_sendmsg() : ectp_skb_alloc() failed.", -1);

		skb->pktlen	= ECTP_PKT_DT_LEN + len;
		skb->pkt	= dt;
		skb->psn	= dt->psn;

		break;

	default:
		return -1;
	}

	return ectp_skb_transmit(skb);
}
