#include <ectp/ectp.h>
#include <ectp/protocol.h>
#include <ectp/structs.h>
#include <ectp/event.h>
#include <ectp/wrapper.h>	/* for err_quit() macro */


/* ------------------------------------------------------------------------- *
 * Implementations of ECTP-3 User Callback Manipulation Functions
 * ------------------------------------------------------------------------- */

void ectp_init_callback( struct ectp_callback *callback )
{
	callback->latejoin_handler = NULL;
	//callback->treejoin_handler = NULL;	/* for join process */
	callback->tokenget_handler = NULL;
}

void ectp_update_callback( struct ectp_callback *dst,
						   struct ectp_callback *src )
{
	memcpy(dst, src, sizeof(struct ectp_callback));
}



/* ------------------------------------------------------------------------- *
 * Implementations of Internal Event Loop & Handlers (Thread Functions).
 * ------------------------------------------------------------------------- */

/**
 * Main Event Loop
 *
 * @param arg	the pointer of ectp_socket{}
 */
void *ectp_main_loop( void *arg )
{
	struct ectp_socket *eso = (struct ectp_socket *) arg;
	struct ectp_handleset *hset = &eso->hset;

	struct timeval *to;
	struct timeval default_to;

	int result;
	fd_set backup_fds, read_fds;
	FD_ZERO(&backup_fds);
	FD_SET(eso->m_sd, &backup_fds);
	FD_SET(eso->u_sd, &backup_fds);

	while ( hset->main_loop_is_running ) {

		read_fds = backup_fds;

		if ( !(to = ectp_timerlist_peek_timeout(&eso->timer_list)) ) {
			default_to.tv_sec	= ECTP_MAINLOOP_TIMEOUT_SEC;
			default_to.tv_usec	= ECTP_MAINLOOP_TIMEOUT_USEC;
			to = &default_to;
		}

		result = select( ((eso->m_sd > eso->u_sd) ? eso->m_sd : eso->u_sd) + 1,
						 &read_fds, NULL, NULL, to );

		switch ( result ) {
			case -1:
				result = 0;
				err_msg("ectp_event_loop() : select() failed.");
				break;
			case 0:
				ectp_timer_dispatch(eso);
				break;
			default:
				ectp_event_dispatch(eso, &read_fds);
				break;
		}
	} /* while() */

	return NULL;
}

void *ectp_join_handler( void *arg )
{
	struct ectp_socket *eso = (struct ectp_socket *) arg;
	struct ectp_endpoint *ep = eso->ep;
	struct ectp_handleset *hset = &eso->hset;

	struct ectp_eventitem *e;

	while ( hset->join_handler_is_running ) {

		e = ectp_event_deq(&eso->joinq);

		if ( !e ) {
			err_msg("ectp_join_handler() : NULL join event occurs. [SKIP]");
			continue;
		}

		/* TO only */
		else if ( e->type == ECTP_REQ_LATE_JOIN ) {
			/* If no handler, deny it. */
			if ( !eso->callback.latejoin_handler ) {
				ectp_ops_deny_latejoin(eso, &e->from);
			}
			else {
				(*(eso->callback.latejoin_handler)) (eso->esd,
						(SS *) &e->from, ectp_addrlen(eso->domain));
			}
		}
		
		/* TO and LO */
		else if ( e->type == ECTP_REQ_TREE_JOIN ) {
			if ( ep->state == ECTP_STATE_ESTABLISHED )
				ectp_ops_accept_treejoin(eso, &e->from);
			else if ( ep->state == ECTP_STATE_CR_SENT ) {
				/* If CCR-triggered phase, there need to verify TJ. */
				ectp_ops_accept_treejoin(eso, &e->from);
			}
			else {
				ectp_ops_deny_treejoin(eso, &e->from);
			}
		} 

		else {
			err_msg("ectp_join_handler() : Event type error.");
		}

		ectp_destroy_event(e);

	} /* while() */

	return NULL;
}

void *ectp_token_handler( void *arg )
{
	struct ectp_socket *eso = (struct ectp_socket *) arg;
	struct ectp_handleset *hset = &eso->hset;
	ectp_token_t	token;

	struct ectp_basehdr *basehdr = NULL;
	struct ectp_eventitem *e = NULL;

	while ( hset->token_handler_is_running ) {

		e = ectp_event_deq(&eso->tokenq);

		if ( !e ) {
			err_msg("ectp_token_handler() : NULL token event occurs. [SKIP]");
			continue;
		}

		else if ( e->type == ECTP_REQ_TOKEN_GET ) {
			/* If no handler, deny it. */
			if ( !eso->callback.tokenget_handler )
				ectp_ops_deny_tokenget(eso, &e->from);
			else {
				(*(eso->callback.tokenget_handler)) (eso->esd,
					(SS *) &e->from, ectp_addrlen(eso->domain));
			}
		}

		else if ( e->type == ECTP_REQ_TOKEN_RET ) {
			basehdr = (struct ectp_basehdr *) e->packet;
			token = basehdr->token_id;
			ectp_ops_accept_tokenreturn(eso, token, &e->from);
		}

		else {
			err_msg("ectp_token_handler() : Event type error.");
		}

		basehdr = NULL;
		ectp_destroy_event(e);

	} /* while() */

	return NULL;
}

void *ectp_event_handler( void *arg )
{
	struct ectp_socket *eso = (struct ectp_socket *) arg;
	struct ectp_endpoint *ep = eso->ep;
	struct ectp_handleset *hset = &eso->hset;

	struct ectp_basehdr *basehdr = NULL;
	struct ectp_eventitem *e = NULL;

	while ( hset->event_handler_is_running ) {

		e = ectp_event_deq(&eso->eventq);

		if ( !e ) {
			err_msg("ectp_event_handler() : NULL event occurs. [SKIP]");
			continue;
		}

		/* According to current implementation,
		 * JR & TJ event is managed in late join procedure. 
		 * SEE: ectp_ops_start_latejoin() */

		else if ( e->type == ECTP_CON_TOKEN_GET ) {
			basehdr = (struct ectp_basehdr *) e->packet;
			ep->token = basehdr->token_id;

			/* awake mtoken_get() */
			ectp_mutex_off(&ep->token_get_mutex);
		}

		else if ( e->type == ECTP_DNY_TOKEN_GET ) {
			ep->token = ECTP_INVALID_TOKEN;

			/* awake mtoken_get() */
			ectp_mutex_off(&ep->token_get_mutex);
		}

		else if ( e->type == ECTP_EXP_TOKEN_GET ) {
			ep->token = ECTP_INVALID_TOKEN;

			/* awake mtoken_get() */
			ectp_mutex_off(&ep->token_get_mutex);
		}

		else if ( e->type == ECTP_CON_TOKEN_RET ) {
			ep->token = ECTP_INVALID_TOKEN;

			/* awake mtoken_return() */
			ectp_mutex_off(&ep->token_ret_mutex);
		}

		else if ( e->type == ECTP_REQ_USER_LEAVE ) {
			ectp_ops_request_userleave( eso, &e->from, 
									ectp_addrlen(eso->domain) );
		}

		else {
			err_msg("ectp_event_handler() : Event type error.");
		}

		basehdr = NULL;
		ectp_destroy_event(e);

	} /* while() */

	return NULL;
}

void ectp_handleset_init( struct ectp_handleset *hset, ectp_role_t role )
{
	hset->main_loop_is_running		= 0;
	hset->event_handler_is_running	= 0;
	hset->join_handler_is_running	= 0;
	hset->token_handler_is_running	= 0;

	hset->main_loop		= ectp_main_loop;
	hset->event_handler	= ectp_event_handler;

	if ( role == ECTP_ROLE_TO ) {
		hset->join_handler	= ectp_join_handler;
		hset->token_handler	= ectp_token_handler;
	}
}



/* ------------------------------------------------------------------------- *
 * Implementations of Internal Event Queue Manipulation Funtions.
 * ------------------------------------------------------------------------- */

struct ectp_eventitem *ectp_create_event( ectp_event_t type,
		ectp_addr_t *from, size_t fromlen, void *packet, __u32 packetlen )
{
	struct ectp_eventitem *_new = NULL;

	if ( !(_new = ectp_ei_alloc()) )
		err_ret("ectp_create_event() : ectp_ei_alloc() failed.", NULL);

	_new->type = type;
	memcpy(&_new->from, from, fromlen);

	if ( !packet || !packetlen )
		return _new;

	_new->packetlen = packetlen;
	_new->packet = malloc(packetlen);
	memcpy(_new->packet, packet, packetlen);

	return _new;
}

void ectp_destroy_event( struct ectp_eventitem *e )
{
	if ( e->packetlen )
		free(e->packet);

	if ( e )
		free(e);

	e = NULL;
}



/**
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_event_initq( struct ectp_eventq *q )
{
	if ( ectp_mutex_init(&q->mutex) < 0 )
		err_ret("ectp_event_initq() : ectp_mutex_init() failed.", -1);

	if ( ectp_sem_init(&q->sem, 0) < 0 ) {
		err_ret("ectp_event_initq() : ectp_sem_init() failed.", -1);
		ectp_mutex_destroy(&q->mutex);
	}

	q->head = q->tail = NULL;
	q->num = 0;

	return 0;
}

/**
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_event_enq( struct ectp_eventq *q, struct ectp_eventitem *_new )
{
	ectp_mutex_on(&q->mutex);
	
	if ( ectp_event_emptyq(q) ) {
		q->head = q->tail = _new;
		_new->next = _new->prev = _new; /* Make doubly linked circular list. */
	} else { /* add_tail() */
		ectp_ei_add(_new, q->tail, q->head); // tail->next : always head.
		q->tail = _new;
	}

	q->num++;
	_new->q = q;

	ectp_mutex_off(&q->mutex);

	ectp_sem_post(&q->sem);

	return 0;
}

struct ectp_eventitem *ectp_event_deq( struct ectp_eventq *q )
{
	struct ectp_eventitem *entry = NULL;

	ectp_sem_wait(&q->sem);	/* blocking! */

	if ( ectp_event_emptyq(q) )		// if ectp_sem_wait() returns,
		return NULL;				// there are one or more items.

	ectp_mutex_on(&q->mutex);

	q->head = q->head->next;
	ectp_ei_del(entry = q->head->prev);

	q->num--;

	/* Although the event queue is empty, 'head' & 'tail' of 'q' point 'entry'.
	 * Therefore we must modify them to NULL. */
	if ( ectp_event_emptyq(q) )
		q->head = q->tail = NULL;

	ectp_mutex_off(&q->mutex);

	return entry;
}



/* ------------------------------------------------------------------------- *
 * IMPEMENTATIONS OF EVENT DISPATCHER
 * ------------------------------------------------------------------------- */

void ectp_event_dispatch(struct ectp_socket *eso, fd_set *fds)
{
	struct ectp_endpoint *ep = eso->ep;
	struct ectp_connection *cn = ep->cn;
	struct ectp_tree *tree = ep->tree;

	ectp_addr_t	from;
	socklen_t fromlen = SA_LEN;

	/* Static buffer size : Base Header Length + MSS */
	__u32 buflen = ECTP_BASEHDR_LEN + cn->mss;
	__u8 buf[buflen];
	int nbytes = -1;

	struct ectp_basehdr *basehdr	= NULL;
	struct ectp_skbuff *skb			= NULL;
	struct ectp_eventitem *event	= NULL;

	if ( FD_ISSET(eso->m_sd, fds) ) {

		nbytes = recvfrom(eso->m_sd, buf, buflen, 0, (SA *) &from, &fromlen);
		basehdr = (struct ectp_basehdr *) buf;

		basehdr->checksum = ectp_checksum((__u16 *) buf, nbytes);
		if ( basehdr->checksum )
			goto discard_event;
		
		if ( basehdr->version != cn->version || basehdr->ct != cn->ct )
			goto discard_event;

		/* Dispatch a multicast packet from TO or the parent(TO/LO). */

		switch ( basehdr->packet_type ) {

		case ECTP_PKT_DT:	/* Multicast DATA from TO. */
			/* Annex B. State Transition Diagrams for ECTP-3
			 * According to Annex B ECTP-3 stack can only send/receive data
			 * in ESTABLISHED state.
			 *
			 * But by using POSIX thread it can be  managed concurrently.
			 * Because we can also send/receive data in JR-RCVD, TGR-SENT,
			 * TGR-RCVD and TRR-SENT states, these states are not necessary.
			 */
			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role != ECTP_ROLE_LO && ep->role != ECTP_ROLE_LE )
				goto discard_event;

			/* Source authentication: from TO ? */
			if ( ectp_addrchk(&tree->to, &from) < 0 )
				goto discard_event;

			if ( !(skb = ectp_skb_alloc(eso, eso->m_sd, &from, fromlen)) )
				err_goto("ectp_event_dispatch() : "
						 "ectp_skb_alloc() failed.", discard_event);
		
			skb->pktlen		= nbytes;
			skb->pkt		= (__u8 *) malloc(nbytes);
			memcpy(skb->pkt, buf, nbytes);
			skb->psn		= basehdr->psn;

			cn->m_ch.rsn = basehdr->psn;		/* rsn update */
			
			if ( cn->ack_buf.vlength == 0 ) {
				/* case : late join */
				cn->m_ch.lsn = cn->m_ch.rsn + 1;

				cn->ack_buf.vlength = 1;
				cn->ack_buf.length = 1;
				cn->ack_bitmapbuf = (unsigned long *)calloc(1, BITMAP_LEN);
				if ( !cn->ack_bitmapbuf ) {
					err_goto("ectp_event_dispatch() : "
							 "alloc of bitmap.", discard_event);
				}
			}

			else {
				/* LSN update */
				if ( cn->m_ch.rsn < cn->m_ch.lsn ) { /* PSN < LSN */
					err_goto("ectp_event_dispatch() : PSN < LSN.", 
							discard_event);
				}
				else if ( cn->m_ch.rsn == cn->m_ch.lsn ) {	/* PSN == LSN */
					if ( cn->ack_buf.vlength == 1 )
						cn->m_ch.lsn++;
					else {
					}
				}
				else {		/* PSN > LSN */
					cn->ack_buf.vlength = cn->m_ch.rsn - cn->m_ch.lsn + 1;

					/* If need, bitmap needs to be extended */
					cn->ack_buf.rowpos = (cn->ack_buf.vlength-1)/32 + 1;

					if ( cn->ack_buf.rowpos > cn->ack_buf.length ) {
						cn->ack_bitmapbuf = 
							(unsigned long *)realloc( cn->ack_bitmapbuf,
											BITMAP_LEN * cn->ack_buf.rowpos );

						if ( !cn->ack_bitmapbuf ) {
							err_goto("ectp_event_dispatch() : "
									 "realloc of bitmap.", discard_event);
						}

						bzero(cn->ack_bitmapbuf + cn->ack_buf.length, 
						 BITMAP_LEN*(cn->ack_buf.rowpos - cn->ack_buf.length));
						cn->ack_buf.length = cn->ack_buf.rowpos;
					}

					if ( !cn->ack_bitmapbuf ) {
						err_goto("ectp_event_dispatch() : "
								 "can't set bitmap.", discard_event);
					}

					bit_set( cn->ack_bitmapbuf, (cn->m_ch.rsn - cn->m_ch.lsn));
				}
			}

			/* Add new soket buffer to the multicast receive buffer */
			ectp_skblist_add(&eso->m_rcvbuf, skb);

			/* ACK Generation */
			if ( (cn->m_ch.rsn % cn->agn) == 
					(unsigned int)(tree->cid % cn->agn) ) {
				ectp_ops_ack(eso);
			}
			
			break;

		case ECTP_PKT_CT:
			printf( "TU - received CT.\n");

			if ( ep->role != ECTP_ROLE_LE )
				goto discard_event;
			
			/*
			 * ECTP-3 API mrecvmsg() return value is 0. 
			 * The mean is Connection Termination.
			 */
			/* Source authentication: from TO ? */
			if ( ectp_addrchk(&tree->to, &from) < 0 )
				goto discard_event;

			if ( !(skb = ectp_skb_alloc(eso, eso->m_sd, &from, fromlen)) )
				err_goto("ectp_event_dispatch() : "
						 "ectp_skb_alloc() failed.", discard_event);
		
			skb->pktlen		= 0;
			skb->pkt		= (__u8 *) malloc(nbytes);
			memcpy(skb->pkt, buf, nbytes);

			/* Add new soket buffer to the multicast receive buffer */
			ectp_skblist_add(&eso->m_rcvbuf, skb);

			/* destroy */
			ectp_ops_end_events(eso);

			if ( ectp_destroy_connection(eso) < 0 )
				err_msg("event_dispatch() : ectp_destroy_connection() failed.");

			if ( ectp_destroy_endpoint(eso->ep) < 0 )
				err_msg("event_dispatch() : ectp_destroy_endpoint() failed.");

			if ( ectp_destroy_socket(eso) < 0 )
				err_msg("event_dispatch() : ectp_destroy_socket() failed.");

			break;

		/* Followings are not implemented in the implementation. */

		case ECTP_PKT_CR:	/* Multicast CR from TO to TUs */
		case ECTP_PKT_RD:	/* Multicast RD from Parent to Children */
		case ECTP_PKT_ND:	/* Multicast ND from TO to TUs */
		case ECTP_PKT_HB:	/* Multicast HB from Parent to Children */
			break;

		/* Followings are unicast-only packet. */

		case ECTP_PKT_CC:
		case ECTP_PKT_TJ:
		case ECTP_PKT_TC:
		case ECTP_PKT_ACK:	
		case ECTP_PKT_HBACK:
		case ECTP_PKT_JR:
		case ECTP_PKT_JC:
		case ECTP_PKT_LR:
		case ECTP_PKT_TGR:
		case ECTP_PKT_TGC:
		case ECTP_PKT_TRR:
		case ECTP_PKT_TRC:
		default:
			break;

		} /* switch ( basehdr->packet_type ) */
	} /* if ( FD_ISSET(eso->m_sd, fds) ) */

	else if ( FD_ISSET(eso->u_sd, fds) ) {
		nbytes = recvfrom(eso->u_sd, buf, buflen, 0, (SA *) &from, &fromlen);
		basehdr = (struct ectp_basehdr *) buf;

		/* checksum */
		basehdr->checksum = ectp_checksum((__u16 *) buf, nbytes);
		if ( basehdr->checksum )
			goto discard_event;
		
		if ( basehdr->version != cn->version || basehdr->ct != cn->ct )
			goto discard_event;

		/* Dispatch an unicast packet. */

		switch ( basehdr->packet_type ) {

		case ECTP_PKT_DT:	/* Unicast DT from SU to TO */

			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role != ECTP_ROLE_TO )
				goto discard_event;

			if ( !(skb = ectp_skb_alloc(eso, eso->u_sd, &from, fromlen)) )
				err_goto("ectp_event_dispatch() : "
						 "ectp_skb_alloc() failed.", discard_event);
		
			skb->pktlen		= nbytes;
			skb->pkt		= (__u8 *) malloc(nbytes);
			memcpy(skb->pkt, buf, nbytes);
			skb->psn		= basehdr->psn;

			/* Add new soket buffer to the unicast receive buffer */
			ectp_skblist_add(&eso->u_rcvbuf, skb);
			break;

		case ECTP_PKT_ACK:	/* Unicast ACK from child(LO/LE) to parent(TO/LE) */
			/* Treate ACK. */
			break;

		case ECTP_PKT_JR:	/* Unicast JR from LE to TO */

			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role != ECTP_ROLE_TO )
				err_goto("ectp_event_dispatch() : "
						 "Only TO can receive JR.", discard_event);

			/* Source authentication is performed by user callback. */

			event = ectp_create_event(
					ECTP_REQ_LATE_JOIN, &from, fromlen, NULL, 0);

			if ( !event )
				err_goto("ectp_event_dispatch() : cannot create JR event.",
						discard_event);

			if ( ectp_event_enq(&eso->joinq, event) < 0 )
				err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
						discard_event);

			break;

		case ECTP_PKT_JC:	/* Unicast JC from TO to LE */

			if ( ep->state != ECTP_STATE_JR_SENT )
				goto discard_event;

			if ( ep->role != ECTP_ROLE_LE )
				err_goto("ectp_event_dispatch() : Only TU can receive JC.",
						discard_event;);

			/* Source authentication: really from TO ? */
			if ( ectp_addrchk(&tree->to, &from) < 0 )
				goto discard_event;

			/* JR is denied by TO (JC with f=0). */
			if ( !basehdr->f ) {
				event = ectp_create_event(
						ECTP_DNY_LATE_JOIN, &from, fromlen, NULL, 0);
			}
			
			/* JR is accepted by TO (JC with f=1). */
			else if ( basehdr->f ) {
				event = ectp_create_event(
						ECTP_CON_LATE_JOIN, &from, fromlen, buf, nbytes);
			}

			if ( !event )
				err_goto("ectp_event_dispatch() : cannot create JC event.",
						discard_event);

			if ( ectp_event_enq(&eso->eventq, event) < 0 )
				err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
						discard_event);
			break;

		case ECTP_PKT_TJ:	/* Unicast TJ from Child(LE/LO) to Parent(LO/TO) */

			if ( ep->state != ECTP_STATE_CR_SENT &&
					ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role != ECTP_ROLE_TO && ep->role != ECTP_ROLE_LO)
				err_goto("ectp_event_dispatch() : TU cannot receive TJ.",
						discard_event);

			event = ectp_create_event(
					ECTP_REQ_TREE_JOIN, &from, fromlen, NULL, 0);

			if ( !event )
				err_goto("ectp_event_dispatch() : cannot create TJ event.",
						discard_event);

			if ( ectp_event_enq(&eso->joinq, event) < 0 )
				err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
						discard_event);
			break;

		case ECTP_PKT_TC:	/* Unicast TC from Parent(TO/LO) to Child(LE/LO) */

			/* Not only LE but also LO must run 'tree-join'. */
			if ( ep->role != ECTP_ROLE_LE && ep->role != ECTP_ROLE_LO )
				err_goto("ectp_event_dispatch() : Only LE/LO can receive TC.",
						discard_event;);

			/* Source authentication: really from Parent ? */
			if ( ectp_addrchk(&tree->parent, &from) < 0 )
				goto discard_event;

			/* Tree Join Confirm triggered by JR->JC->TJ. */
			if ( ep->state == ECTP_STATE_TJ_SENT ) {

				/* TJ is denied by Parent (TC with f=0). */
				if ( !basehdr->f ) {
					event = ectp_create_event(
							ECTP_DNY_TREE_JOIN, &from, fromlen, NULL, 0);
				}
				
				/* TJ is accepted by Parent (TC with f=1). */
				else {
					event = ectp_create_event(
							ECTP_CON_TREE_JOIN, &from, fromlen, buf, nbytes);
				}

				if ( !event )
					err_goto("ectp_event_dispatch() : cannot create TJ event.",
							discard_event);

				if ( ectp_event_enq(&eso->eventq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
							discard_event);
			}

			else if ( ep->state == ECTP_STATE_CR_RCVD ) {
			}

			break;

		case ECTP_PKT_TGR:	/* Unicast TGR from LE/LO to TO */

			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role == ECTP_ROLE_TO ) {

				/* Source authentication is performed by user callback. */
				event = ectp_create_event(
						ECTP_REQ_TOKEN_GET, &from, fromlen, NULL, 0);
				if ( !event )
					err_goto("ectp_event_dispatch() : cannot create TGR event.",
							discard_event);

				if ( ectp_event_enq(&eso->tokenq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_token_enq() failed.",
							discard_event);
			}

			break;

		case ECTP_PKT_TGC:	/* Unicast TGC from LE/LO to TO */
			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			/* Token Get Confirm triggered by TO */
			if ( ep->role == ECTP_ROLE_LE || ep->role == ECTP_ROLE_LO ) {

				/* Source authentication: really from TO ? */
				if ( ectp_addrchk(&tree->to, &from) < 0 )
					goto discard_event;

				/* TGR is denied by TO (TGC with f=0). */
				if ( !basehdr->f ) {
					event = ectp_create_event(
							ECTP_DNY_TOKEN_GET, &from, fromlen, NULL, 0);
				}
				
				/* TGR is accepted by TO (TGC with f=1). */
				else {
					event = ectp_create_event(
							ECTP_CON_TOKEN_GET, &from, fromlen, buf, nbytes);
				}

				if ( !event )
					err_goto("ectp_event_dispatch() : "
							 "ectp_create_event() failed.", discard_event);

				if ( ectp_event_enq(&eso->eventq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
							discard_event);
			}
			break;

		case ECTP_PKT_TRR:	/* Unicast TRR from SU to TO */
			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			if ( ep->role == ECTP_ROLE_TO ) {

				/* Source authentication is performed by user callback. */

				event = ectp_create_event(
						ECTP_REQ_TOKEN_RET, &from, fromlen, buf, nbytes);
				if ( !event )
					err_goto("ectp_event_dispatch() : cannot create TRR event.",
							discard_event);

				if ( ectp_event_enq(&eso->tokenq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_token_enq() failed.",
							discard_event);

				printf( "Token Return Request\n" );

			}
			break;

		case ECTP_PKT_TRC:	/* Unicast TRC from TO to SU */

			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			/* Token Return Confirm triggered by TO */
			if ( ep->role == ECTP_ROLE_LE || ep->role == ECTP_ROLE_LO ) {

				/* Source authentication: really from TO ? */
				if ( ectp_addrchk(&tree->to, &from) < 0 )
					goto discard_event;

				/* TRR is denied by TO (TRC with f=0). */
				if ( !basehdr->f ) {
//					event = ectp_create_event(
//							ECTP_DNY_TOKEN_RET, &from, fromlen, NULL, 0);
				}
				
				/* TRR is accepted by TO (TRC with f=1). */
				else {
					event = ectp_create_event(
							ECTP_CON_TOKEN_RET, &from, fromlen, buf, nbytes);
				}

				if ( !event )
					err_goto("ectp_event_dispatch() : "
							 "ectp_create_event() failed.", discard_event);

				if ( ectp_event_enq(&eso->eventq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
							discard_event);
			}

			break;

		case ECTP_PKT_LR:	/* Unicast LR from Child/Parent to Parent/Child */

			if ( ep->state != ECTP_STATE_ESTABLISHED )
				goto discard_event;

			/* User Leave triggered by Child */
			if ( ep->role == ECTP_ROLE_TO ) {
				event = ectp_create_event(
						ECTP_REQ_USER_LEAVE, &from, fromlen, buf, nbytes);
				printf( "User Leave\n" );

				if ( !event )
					err_goto("ectp_event_dispatch() : "
							 "ectp_create_event() failed.", discard_event);

				if ( ectp_event_enq(&eso->eventq, event) < 0 )
					err_goto("ectp_event_dispatch() : ectp_event_enq() failed.",
							discard_event);
			}

			else if ( ep->role == ECTP_ROLE_LO ) {
				/* From TO => same as the case by Parent. */
				/* From a valid child => same as the case by Child. */
			}

			/* User Leave triggered by Parent */
			else if ( ep->role == ECTP_ROLE_LE ) {
				/* User Leave by Parent */
			}

			/* Enters SHUTDOWN-PENDING */
			break;

		/* Followings are not implemented in this implementation. */

		case ECTP_PKT_CC:	/* Unicast CC from LE(TU) to TO */
		case ECTP_PKT_RD:	/* Unicast RD from SU to TO */
		case ECTP_PKT_HB:	/* Unicast HB from SU to TO */
		case ECTP_PKT_HBACK:/* Unicast HB from TO/parent to SU/child */
			break;

		/* Followings are multicast-only packets. */

		case ECTP_PKT_CR:
		case ECTP_PKT_ND:	
		case ECTP_PKT_CT:
		default:
			break;
		} 
	} /* else if ( FD_ISSET(eso->u_sd, fds) ) */

	return;

discard_event:
	memset(buf, 0, buflen);
	if ( event )
		ectp_destroy_event(event);

} /* ectp_event_dispatch() */


