/* Enhanced Communication Transport Protocol Part-3 Implementation
 *
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * ectp_subr.c
 * This file is the implementation of the ECTP-3 internal functions for
 * Sockets  API.
 * 
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers <saintpaul1978@mail.knu.ac.kr>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 */


#include <ectp/types.h>
#include <ectp/wrapper.h>
#include <ectp/protocol.h>
#include <ectp/structs.h>

#include <string.h>			/* for memset() */
#include <stdlib.h>			/* for malloc(),calloc(),free() */
#include <unistd.h>			/* for close() */


/* ------------------------------------------------------------------------- *
 * Definitions for ECTP-3 Global Variables.
 * ------------------------------------------------------------------------- */

/**
 * ECTP-3 Global Variables
 */
struct ectp_globals		*ectp_global = NULL;

/**
 * Socket Pointer that is used to Current Process.
 */
struct ectp_socket		*ectp_current = NULL;

/**
 * ectp_create_globals() create the global variable structure.
 * - On success a pointer to the allocated memory is returned.
 * - Otherwise NULL is returned.
 */
struct ectp_globals *ectp_create_globals()
{
	struct ectp_globals *_new = NULL;

	_new = (struct ectp_globals *) calloc(1, sizeof(struct ectp_globals));
	if ( !_new )
		err_ret("ectp_create_globals() : calloc() failed.", NULL);

	if ( ectp_mutex_init(&_new->mutex) < 0 ) {
		free(_new);
		err_ret("ectp_create_globals() : ectp_mutex_init() failed.", NULL);
	}

	return _new;
}


void ectp_destroy_globals( struct ectp_globals *global )
{
	/* Not yet. */
}



/* ------------------------------------------------------------------------- *
 * Implementation of Functions for the ECTP-3 Socket Buffer Structure
 * ------------------------------------------------------------------------- */

/**
 * ectp_skblist_init() makes ready to use ECTP-3 socket buffer as a
 * input/output queue.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_skblist_init( struct ectp_skblist *list, __u32 max_length )
{
	if ( ectp_mutex_init(&list->mutex) < 0 ) {
		err_ret("ectp_skblist_init() : ectp_mutex_init() failed.", -1);
	}

	if ( ectp_sem_init(&list->sem, 0) < 0 ) {
		err_ret("ectp_skblist_init() : ectp_sem_init() failed.", -1);
		ectp_mutex_destroy(&list->mutex);
	}

	list->head = list->tail = NULL;
	list->num = list->length = 0;
	list->max_length = max_length;

	return 0;
}

/**
 * ectp_skblist_add() appends a new socket buffer to the given list.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_skblist_add( struct ectp_skblist *list,
									 struct ectp_skbuff *_new )
{
	ectp_mutex_on(&list->mutex);

	if ( ectp_skblist_empty(list) ) {
		list->head = list->tail = _new;
		_new->next = _new->prev = _new;	/* Make doubly linked circular list. */
	} else {
		/* check the space that can be used. */
		if ( ectp_skblist_spare(list) < _new->pktlen ) {
			ectp_mutex_off(&list->mutex);
			err_ret("ectp_skblist_add() : cannot add anymore.", -1);
		}

		/* add_tail() */
		ectp_skb_add(_new, list->tail, list->head);
		list->tail = _new;
	}

	list->num++;
	list->length += _new->pktlen;
	_new->list = list;

	ectp_mutex_off(&list->mutex);

	ectp_sem_post(&list->sem);

	return 0;
}

/**
 * ectp_skblist_del() deletes head node and return it.
 * -	If 'skb' is NULL, the head node of 'list' is deleted and returned.
 * -	If 'skb' is specified socket buffer, it is deleted from the list.
 */
struct ectp_skbuff *ectp_skblist_del( struct ectp_skblist *list,
									  struct ectp_skbuff  *skb )
{
	struct ectp_skbuff *entry = NULL;

	/* POSIX semaphore is safety on signal? */
	ectp_sem_wait(&list->sem);

	ectp_mutex_on(&list->mutex);

	if ( !skb ) {
		list->head = list->head->next;
		ectp_skb_unlink(entry = list->head->prev);
		list->num--;
		list->length -= entry->pktlen;
	} else {
		if ( skb == list->head )
			list->head = skb->next;
		else if ( skb == list->tail )
			list->tail = skb->prev;
		ectp_skb_unlink(entry = skb);
	}

	/* Although the list is empty, 'list->head' and 'list->tail' point 'entry'.
	 * Therefore we must modify them to NULL.
	 */
	if ( ectp_skblist_empty(list) ) {
		list->head = list->tail = NULL;
	}

	if ( ectp_mutex_off(&list->mutex) < 0 )
		err_ret("ectp_skblist_del() : ectp_mutex_off() failed.", entry);

	return entry;
}



/* -------------------------------------------------------------------------- *
 * Implementation of ECTP-3 Internal Functions for Manipulating Socket.
 * -------------------------------------------------------------------------- */

/**
 * ectp_create_socket() creates ECTP-3 socket structure for communications.
 * - On success, a pointer of a new ECTP-3 socket is returned.
 *   Otherwise, NULL is returned.
 */

/* ADVANCED SUBJECT:
 * To support multiple ECTP-3 endpoint, we need to append following procedure
 * when creating a new socket{}:
 *
 * 	1.	Check the defined UNIX domain socket.
 * 	2.	If not defined, invoke the ECTP-3 user space kernel by using exec().
 * 	3.	If defined, just request the kernel to create a new socket{}.
 */

struct ectp_socket *ectp_create_socket( int domain, ectp_role_t role )
{
	struct ectp_socket *_new = NULL;

	/* Initialize Global Variables. */
	if ( !ectp_global )
		ectp_global = ectp_create_globals();

	/* Check boundary condition. */
	if ( ectp_global->nsock >= ECTP_MAX_SOCKETS )
		err_ret("ectp_create_socket() : cannot create anymore.", NULL);

	/* Allocate memory for a new socket. */
	if ( !(_new = ectp_sk_alloc()) )
		err_ret("ectp_create_socket() : ectp_sk_alloc() failed.", NULL);

	/* Initialize the internal fields and open new UDP socket. */
	if ( ectp_sk_init(_new, domain, role) < 0 ) {
		ectp_sk_free(_new);
		err_ret("ectp_create_socket() : ectp_sk_init() failed.", NULL);
	}

	/* Add current socket to a global array/list. */
	//ectp_sk_addlist(_new);

	ectp_mutex_on(&ectp_global->mutex);

	ectp_global->socks[ectp_global->nsock++] = _new;

	if ( ectp_mutex_off(&ectp_global->mutex) < 0 ) {
		ectp_sk_free(_new);
		ectp_global->nsock--;
		err_ret("ectp_create_socket() : ectp_mutex_off() failed.", NULL);
	}

	/* Set the current pointer. */
	ectp_current = _new;

	return _new;
}

/**
 * ectp_destroy_socket() destroies ECTP-3 endpoint.
 *   This function deallocates all the memories that are allocated for protocol
 *   operation. After destroy, the user cannot use ECTP-3 APIs anymore.
 */
int ectp_destroy_socket( struct ectp_socket *eso )
{
	ectp_sk_free(eso);
	close(eso->m_sd);
	close(eso->u_sd);

	return 0;
}

/**
 * ectp_sk_alloc() allocates new memory for given ECTP-3 socket pointer.
 *
 * RETURN:
 * 	-	On success ~~ is returned.
 * 	-	Otherwise NULL is returned.
 */
struct ectp_socket *ectp_sk_alloc()
{
	struct ectp_socket *eso = NULL;

	eso = (struct ectp_socket *) calloc(1, ECTP_SOCKET_LEN);
	if ( !eso )
		err_ret("ectp_sk_alloc() : calloc() failed.", NULL);

	return eso;
}

/**
 * ectp_sk_free() deallocates memory of the given ECTP-3 socket.
 */
void ectp_sk_free( struct ectp_socket *eso )
{
	if ( eso )
		free(eso);

	eso = NULL;
}

/**
 * ectp_sk_init() initializes each field of the given ECTP-3 socket. It also
 * open a new UDP socket for communications.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_sk_init( struct ectp_socket *eso, int domain, ectp_role_t role )
{
	/* Open two UDP sockets. */

	if ( (eso->m_sd = socket(domain, SOCK_DGRAM, IPPROTO_UDP)) < 0 ) {
		close(eso->m_sd);
		err_ret("ectp_sk_init() : multi channel - socket() failed.", -1);
	}

	if ( (eso->u_sd = socket(domain, SOCK_DGRAM, IPPROTO_UDP)) < 0 ) {
		close(eso->u_sd);
		err_ret("ectp_sk_init() : uni channel - socket() failed.", -1);
	}

	/* NOTE:
	 * Currently eso->esd use array index.
	 * It will be needed to append socketlist{} and hash function. */
	eso->esd = ectp_global->nsock;

	eso->domain = domain;
	eso->ep = NULL;
	ectp_init_callback(&eso->callback);

	/* Initialize timer list. */
	if ( ectp_timerlist_init(&eso->timer_list) < 0 )
		err_goto("ectp_sk_init() : cannot initialize timer list.", close_udp);

	/* Initialize event list. */

	if ( role == ECTP_ROLE_TO || role == ECTP_ROLE_LO ) {
		if ( ectp_event_initq(&eso->joinq) < 0 )
			err_goto("ectp_sk_init() : "
					 "cannot initialize join queue.", close_udp);
	}

	if ( role == ECTP_ROLE_TO ) {
		if ( ectp_event_initq(&eso->tokenq) < 0 )
			err_goto("ectp_sk_init() : "
					 "cannot initialize tokenq queue.", close_udp);
	}

	/* If TU, only one event queue is needed. */
	if ( ectp_event_initq(&eso->eventq) < 0 )
		err_goto("ectp_sk_init() : cannot initialize event queue.", close_udp);

	/* Initialize Event Handler. */
	ectp_handleset_init(&eso->hset, role);

	eso->max_sndbuf = ECTP_PARAM_SNDBUF_DFL;
	eso->max_rcvbuf = ECTP_PARAM_RCVBUF_DFL;

	/* Initialize the send buffer. */

	if ( role == ECTP_ROLE_TO || role == ECTP_ROLE_LO )
		if ( ectp_skblist_init(&eso->m_sndbuf, ECTP_PARAM_SNDBUF_DFL) < 0 )
			err_goto("ectp_sk_init() : ectp_skblist_init() failed.", close_udp);
	
	/* If TU doesn't send any multicast DATA packet, TU doesn't need send
	 * buffer for multicast DATA packet anymore. */
	if ( ectp_skblist_init(&eso->u_sndbuf, ECTP_PARAM_SNDBUF_DFL) < 0 )
		err_goto("ectp_sk_init() : ectp_skblist_init() failed.", close_udp);


	/* Initialize the receive buffer. */

	if ( role == ECTP_ROLE_LE || role == ECTP_ROLE_LO )
		if ( ectp_skblist_init(&eso->m_rcvbuf, ECTP_PARAM_RCVBUF_DFL) < 0 )
			err_goto("ectp_sk_init() : ectp_skblist_init() failed.", close_udp);

	/* If TO doesn't send any unicast DATA packet, TU doesn't need receive
	 * buffer for unicast DATA packet anymore. */
	if ( ectp_skblist_init(&eso->u_rcvbuf, ECTP_PARAM_RCVBUF_DFL) < 0 )
		err_goto("ectp_sk_init() : ectp_skblist_init() failed.", close_udp);


	return 0;

close_udp:
	close(eso->m_sd);
	close(eso->u_sd);
	return -1;
}


/*
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_sk_bind( struct ectp_socket *eso,
				  SS *local_addr, SS *group_addr )
{
	struct ectp_endpoint *ep = eso->ep;

	struct ip_mreq mreq;

	int addrlen;
	int result;
	const int on = 1;

	/* Check the domain */
	if ( eso->domain != local_addr->ss_family ||
			eso->domain != group_addr->ss_family ) {
		err_ret("ectp_sk_bind() : address families are different to "
				"socket's one.", -1);
	}

	/* If socket domain and address families match, ... */

	/* Binding & internal UDP binding for Unicast Channel. */

	addrlen = ectp_addrlen(eso->domain);
	memcpy(&eso->local_addr, local_addr, addrlen);
	memcpy(&eso->group_addr, group_addr, addrlen);


	if ( ep->role == ECTP_ROLE_TO ) {

		setsockopt(eso->u_sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

		/* TO : unicast binding using local address & group port. */
		if ( eso->domain == AF_INET ) {
			if ( bind(eso->u_sd, (SA *) &eso->local_addr, addrlen) < 0 ) {
				err_ret("ectp_sk_bind() : bind() failed.", -1);
			}
		}
		else if ( eso->domain == AF_INET6 ) {
			/* AF_INET6 */
		}
	}

	else if ( ep->role == ECTP_ROLE_LE ) { 

		setsockopt(eso->m_sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

		/* LE : multicast binding using group address & group port. */
		if ( eso->domain == AF_INET ) {
			if ( bind(eso->m_sd, (SA *) &eso->group_addr, addrlen) < 0 ) {
				err_ret("ectp_sk_bind() : bind() failed.", -1);
			}

			mreq.imr_multiaddr.s_addr = eso->group_addr.sa.sin_addr.s_addr;
			mreq.imr_interface.s_addr = htonl(INADDR_ANY);

			result  = setsockopt( eso->m_sd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
							&mreq, sizeof(struct ip_mreq) );
			if ( result )
				err_ret("ectp_sk_bind() : setsockopt() failed.", -1);

		}
		
		else if ( eso->domain == AF_INET6 ) {
			/* AF_INET6 */
		}
	}

	else if ( ep->role == ECTP_ROLE_LO ) {
		/* LO : both multicast binding and unicast binding */
		if ( eso->domain == AF_INET ) {
		}
		else if ( eso->domain == AF_INET6 ) {
		}
	}

	return 0;
}




/* ------------------------------------------------------------------------- *
 * Implementation of Functions for the ECTP-3 Communication Endpoint
 * ------------------------------------------------------------------------- */

/**
 * ectp_create_endpoint() creates ECTP-3 Protocol Endpoint for the given socket
 * structure.
 *
 * RETURN:
 * 		On success 0 is returned.
 * 		Otherwise -1 is returned.
 */
int ectp_create_endpoint( struct ectp_socket *eso, ectp_role_t role )
{
	struct ectp_endpoint *ep = NULL;
	struct ectp_connection *cn = NULL;

	/* Allocate memory for a new endpoint. */
	if ( !(ep = ectp_ep_alloc()) < 0 )
		err_ret("ectp_create_endpoint() : ectp_ep_alloc() failed.", -1);

	ep->eso = eso;

	/* Allocate memory for a new connection */
	if ( !(cn = ectp_cn_alloc()) < 0 )
		err_ret("ectp_create_endpoint() : ectp_cn_alloc() failed.", -1);

	/* Init. connection */
	ectp_cn_init(eso, cn);

	ep->cn = cn;

	/* ep->tree will be allocated when connection/join is made. */
	ep->tree = NULL;

	ep->state = ECTP_STATE_CLOSED;
	ep->role = role;

	/* Token ID '0' is reserved to use of TO. */
	if ( role == ECTP_ROLE_TO ) {
		ep->token = ECTP_TOKEN_TO;
		if ( !(ep->tokenset = ectp_create_tokenset()) ) {
			ectp_ep_free(ep);
			err_ret("ectp_create_endpoint() : "
					"ectp_create_tokenset() failed.", -1);
		}
	} else {
		ep->token = ECTP_INVALID_TOKEN;
		ep->tokenset = NULL;
	}

	ectp_mutex_init(&ep->token_get_mutex);

	/* Make the resource empty. This job is to blocking operation of
	 * mtoken_get(). */
	ectp_mutex_on(&ep->token_get_mutex);

	eso->ep = ep;

	return 0;
}

int ectp_destroy_endpoint( struct ectp_endpoint *ep )
{
	ectp_ep_free(ep);

	return 0;
}

/**
 * ectp_ep_alloc() allocates new memory for given ECTP-3 endpoint.
 *
 * RETURN:
 * 		On success ~~ is returned.
 * 		Otherwise NULL is returned.
 */
struct ectp_endpoint *ectp_ep_alloc()
{
	struct ectp_endpoint *ep;
	ep = (struct ectp_endpoint *) calloc(1, ECTP_ENDPOINT_LEN);
	if ( !ep )
		err_ret("ectp_ep_alloc() : calloc() failed.", NULL);

	return ep;
}

/**
 * ectp_ep_free() deallocates memory of the given ECTP-3 endpoint.
 */
void ectp_ep_free( struct ectp_endpoint *ep )
{
	if ( ep )
		free(ep);

	ep = NULL;
}





/* ------------------------------------------------------------------------- *
 * Implementation of Functions for the ECTP-3 Connection Structure. 
 * ------------------------------------------------------------------------- */


/** TO/TU */
int ectp_create_connection( struct ectp_socket *eso,
						    SS *local_addr, SS *group_addr, int cct)
{
	/* bind given addresses to the socket & UDP binding for Unicast Channel. */
	if ( ectp_sk_bind(eso, local_addr, group_addr) < 0 )
		err_ret("ectp_create_connection() : ectp_sk_bind() failed.", -1);

	/* CR Wait Time setting */
	eso->ep->cn->ctime = cct;

	return 0;
}

int ectp_destroy_connection( struct ectp_socket *eso )
{
	ectp_cn_free(eso->ep->cn);
	return 0;
}

/**
 * ectp_cn_alloc() allocates new memory for given ECTP-3 connection.
 *
 * RETURN:
 * 		On success a new pointer of a ECTP-3 connection is returned.
 * 		Otherwise NULL is returned.
 */
struct ectp_connection *ectp_cn_alloc()
{
	struct ectp_connection *_new = NULL;

	_new = (struct ectp_connection *) calloc(1, ECTP_CONNECTION_LEN);
	if ( !_new )
		err_ret("ectp_ep_alloc() : calloc() failed.", NULL);

	return _new;
}

/**
 * ectp_cn_free() deallocates memory of the given ECTP-3 connection.
 */
void ectp_cn_free( struct ectp_connection *cn )
{
	if ( cn )
		free(cn);

	cn = NULL;
}

void ectp_cn_init( struct ectp_socket *eso,	struct ectp_connection *cn )
{
	cn->ep			= eso->ep;

	cn->version 	= ECTP_PARAM_VER_DFL;	/* 0 */
	cn->ct 			= ECTP_PARAM_CT_DFL;	/* 2 */

	cn->tco			= ECTP_PARAM_TCO_DFL;	/* 0 */
	cn->agn 		= ECTP_PARAM_AGN_DFL;	/* 2 */
	cn->mss 		= ECTP_PARAM_MSS_DFL;	/* 1024 */
	cn->fdn 		= ECTP_PARAM_FDN_DFL;	/* 3 */

	cn->connected	= 0;
	cn->ctime		= 3;	/* optional value */

	cn->m_ch.rxt	= ECTP_PARAM_RXT_DFL;	/* 3 [secs] */
	cn->m_ch.hgt	= ECTP_PARAM_HGT_DFL;	/* 9 [secs] */
	cn->m_ch.isn	= ECTP_PARAM_ISN_DFL;	/* 0 */
	cn->m_ch.lsn	= ECTP_PARAM_LSN_DFL;	/* 0 */
	cn->m_ch.rsn	= 0;
	cn->m_ch.ssn	= 0;

	cn->u_ch.rxt	= ECTP_PARAM_RXT_DFL;	/* 3 [secs] */
	cn->u_ch.hgt	= ECTP_PARAM_HGT_DFL;	/* 9 [secs] */
	cn->u_ch.isn	= ECTP_PARAM_ISN_DFL;	/* 0 */
	cn->u_ch.lsn	= ECTP_PARAM_LSN_DFL;	/* 0 */
	cn->u_ch.rsn	= 0;
	cn->u_ch.ssn	= 0;

	cn->ack_buf.next = ECTP_EE_NOELEMENT;
	cn->ack_buf.length = 0;
	cn->ack_buf.vlength = 0;
	cn->ack_buf.rowpos = 0;

}





/* ------------------------------------------------------------------------- *
 * Implementation of Functions for the ECTP-3 Control Tree Structure. 
 * ------------------------------------------------------------------------- */

int ectp_create_tree( struct ectp_socket *eso )
{
	struct ectp_tree *_new = NULL;

	if ( !(_new= (struct ectp_tree *) ectp_tree_alloc()) )
		err_ret("ectp_create_tree() : ectp_tree_alloc() failed.", -1);

	if ( ectp_tree_init(eso, _new, eso->ep->role) < 0 ) {
		ectp_tree_free(_new);
		err_ret("ectp_create_tree() : ectp_tree_init() failed.", -1);
	}

	eso->ep->tree = _new;

	return 0;
}

int ectp_destroy_tree( struct ectp_socket *eso )
{
	ectp_tree_free(eso->ep->tree);
	
	return 0;
}

struct ectp_tree *ectp_tree_alloc()
{
	struct ectp_tree *_new;

	_new = (struct ectp_tree *) calloc(1, ECTP_TREE_LEN);
	if ( !_new )
		err_ret("ectp_tree_alloc() : calloc() failed.", NULL);

	return _new;
}

void ectp_tree_free( struct ectp_tree *tree )
{
	if ( ectp_destroy_childset(tree->childset) < 0 )
		err_msg("ectp_tree_free() : ectp_destroy_childset() failed.");

	if ( tree )
		free(tree);

	tree = NULL;
}

int ectp_tree_init( struct ectp_socket *eso,
		struct ectp_tree *tree, ectp_role_t role )
{
	tree->ep = eso->ep;
	tree->cid = ECTP_INVALID_CID; /* 0, it would be ignored.  */

	if ( role == ECTP_ROLE_TO ) {
		tree->level		= 0;
		tree->adn		= 0;
		if ( !(tree->childset = ectp_create_childset()) ) {
			err_ret("ectp_create_tree() : "
					"ectp_create_childset() failed.", -1);
		}
		memcpy(&tree->to, &eso->local_addr, ECTP_ADDR_LEN);
		/* tree->parent is not used. */
	} else if ( role == ECTP_ROLE_LE ) {
		tree->level		= 1;
		tree->adn		= 1;
		tree->childset	= NULL;
		memcpy(&tree->to, &eso->local_addr, ECTP_ADDR_LEN);
		/* tree->parent are set, when TU receives 'CR' or 'JC'. */
	} else if ( role == ECTP_ROLE_LO ) {
		/* Case LO */
	}

	return 0;
}
