
#include <ectp/ectp.h>
#include <ectp/protocol.h>
#include <ectp/structs.h>
#include <ectp/event.h>
#include <ectp/wrapper.h>	/* for err_quit() macro */

#include <sys/time.h>	/* gettimeofday() */
#include <time.h>




/* ------------------------------------------------------------------------- *
 * Implementations of ECTP-3 Timer Event (Alarm) Manipulation Functions.
 * ------------------------------------------------------------------------- */

void ectp_ti_free(struct ectp_timeritem *alarm)
{
	if ( alarm->skb ) {
		ectp_skblist_del(alarm->skb->list, alarm->skb);
		ectp_skb_free(alarm->skb);
	}

	free(alarm);
	alarm = NULL;
}

int ectp_timerlist_init( struct ectp_timerlist *list )
{
	if ( ectp_mutex_init(&list->mutex) < 0 )
		err_ret("ectp_timerlist_init() : ectp_mutex_init() failed.", -1);

	list->head = list->tail = NULL;
	list->num  = 0;

	return 0;
}

/** 
 * Add sorted.
 */
int ectp_timerlist_add( struct ectp_timerlist *list,
					    struct ectp_timeritem *_new )
{
	struct ectp_timeritem *t;
	__u32 num = list->num;

	ectp_mutex_on(&list->mutex);

	if ( ectp_timerlist_empty(list) ) {	/* Is list empty ? */
		list->head = list->tail = _new;
		_new->next = _new->prev = _new; /* Make doubly linked circular list. */
	} else {
		t = list->tail;
		while ( num > 0 ) {
			if ( (t->action_time.tv_sec < _new->action_time.tv_sec) ||
					((t->action_time.tv_sec == _new->action_time.tv_sec) &&
					 (t->action_time.tv_usec < _new->action_time.tv_usec)) ) {
				break;
			}
			t = t->prev;
			num--;
		}

		if ( num == list->num ) {/* _new has the largest action_time value */
			ectp_ti_add(_new, list->tail, list->head);
			list->tail = _new;
		} else if ( !num ) {	 /* _new has the smallest action_time value */
			ectp_ti_add(_new, list->tail, list->head);
			list->head = _new;
		} else {
			ectp_ti_add(_new, t, t->next);
		}

	}

	list->num++;
	_new->list = list;

	if ( ectp_mutex_off(&list->mutex) < 0 ) {
		ectp_ti_unlink(_new);
		err_ret("ectp_timerlist_add() : ectp_mutex_off() failed.\n"
				"                     : Check failed timer in the list", -1);
	}

	return 0;
}

/**
 * If alarm is NULL, the head is deleted from the list and returned.
 * If alarm is specified timer item, it is deleted from the list.
 *
 * RETURN:
 * 		If the list is empty, NULL is returned.
 * 		Otherwise the pointer of deleted timer item is returned.
 * 		(If you specify the alarm, it will be unlinked from the list and
 * 		returned.)
 */
struct ectp_timeritem *ectp_timerlist_del( struct ectp_timerlist *list,
										   struct ectp_timeritem *alarm )
{
	struct ectp_timeritem *entry = NULL;

	if ( ectp_timerlist_empty(list) )
		err_ret("ectp_timerlist_del() : timer list is empty.", NULL);

	ectp_mutex_on(&list->mutex);

	/* blocking ... */

	if ( !alarm ) {
		list->head = list->head->next;
		ectp_ti_unlink(entry = list->head->prev);
		list->num--;
	} else {
		/* If alarm is head, ... */
		if ( alarm == list->head )	
			list->head = alarm->next;

		/* If alarm is tail, ... */
		else if ( alarm == list->tail )
			list->tail = alarm->prev;

		ectp_ti_unlink( entry = alarm );
		list->num--;
	}

	/* Although the timer list is empty, 'head'&'tail' of 'list' point 'entry'.
	 * Therefore we must modify them to NULL. */
	if ( ectp_timerlist_empty(list) )
		list->head = list->tail = NULL;

	if ( ectp_mutex_off(&list->mutex) < 0 )
		err_ret("ectp_timerlist_del() : ectp_mutex_off() failed.", entry);

	return entry;
}





/* ------------------------------------------------------------------------- *
 * IMPEMENTATIONS OF TIMER DISPATCHER
 * ------------------------------------------------------------------------- */

void ectp_timer_dispatch( struct ectp_socket *eso )
{
	struct ectp_endpoint *ep = eso->ep;
	struct ectp_connection *cn = ep->cn;

	struct timeval *to = NULL;
	struct ectp_timeritem *timer= NULL;
	struct ectp_skbuff *skb = NULL;
	struct ectp_basehdr *hdr = NULL;

	struct ectp_eventitem *e = NULL;
	
	/* If the timer list is empty, just return. */
	if ( !(to = ectp_timerlist_peek_timeout(&eso->timer_list)) )
		return;

	/* If not expired, just return. */
	if ( to->tv_sec != 0 || to->tv_usec != 0 )
		return;

	/* If expired, ... */

	timer = ectp_timerlist_del(&eso->timer_list, NULL);
	if ( !timer )
		return;

	skb = timer->skb;
	hdr = (struct ectp_basehdr *) skb->pkt;

	switch ( timer->type ) {

	case ECTP_TIMER_RXT:
		/* Check if the retransmission counter does exceed maximum retransmit
		 * number that is the same value as local fdn.
		 */
		if ( ++skb->rtxcnt == cn->fdn ) {

			switch ( ep->state ) {

			case ECTP_STATE_JR_SENT:
				e = ectp_create_event(ECTP_EXP_LATE_JOIN,
						&skb->addr, skb->addrlen, NULL, 0);
				if ( !e ) {
					err_msg("ectp_timer_dispatch() : "
							"ectp_create_event() failed.");
					break;
				}

				if ( ectp_event_enq(&eso->eventq, e) < 0 ) {
					err_msg("ectp_timer_dispatch() : ectp_event_enq() failed.");
					ectp_destroy_event(e);
				}
				break;

			case ECTP_STATE_TJ_SENT:
				e = ectp_create_event(ECTP_EXP_TREE_JOIN,
						&skb->addr, skb->addrlen, NULL, 0);
				if ( !e ) {
					err_msg("ectp_timer_dispatch() : "
							"ectp_create_event() failed.");
					break;
				}

				if ( ectp_event_enq(&eso->eventq, e) < 0 ) {
					err_msg("ectp_timer_dispatch() : ectp_event_enq() failed.");
					ectp_destroy_event(e);
				}
				break;

			case ECTP_STATE_TGR_SENT:
				e = ectp_create_event(ECTP_EXP_TOKEN_GET,
						&skb->addr, skb->addrlen, NULL, 0);
				if ( !e ) {
					err_msg("ectp_timer_dispatch() : "
							"ectp_create_event() failed.");
					break;
				}

				if ( ectp_event_enq(&eso->eventq, e) < 0 ) {
					err_msg("ectp_timer_dispatch() : ectp_event_enq() failed.");
					ectp_destroy_event(e);
				}
				break;

			case ECTP_STATE_TRR_SENT:
				e = ectp_create_event(ECTP_EXP_TOKEN_RET,
						&skb->addr, skb->addrlen, NULL, 0);
				break;
			default:
				break;
			} /* switch ( ep->state ) */
			
#ifdef	_ECTP_DEBUG
			err_goto("DEBUG:  _timer_dispatch() : cannot retransmit any more.",
					discard_timer);
#else
			goto discard_timer;
#endif
		} /* if ( ++skb->rtxcnt == cn->fdn ) { */

		/* If DT, convert it into RD. */
		if ( hdr->packet_type == ECTP_PKT_DT ) {
			hdr->packet_type = ECTP_PKT_RD;
			hdr->checksum = 0;
			hdr->checksum = ectp_checksum((__u16 *) skb->pkt, skb->pktlen);
		}

		/* Recalculate the action timer of 'timer' and move it to the end of
		 * the its list.
		 */ 

		if ( ectp_ti_update(timer) < 0 )
			err_goto("ectp_timer_dispatch() : ectp_ti_update() failed.",
					discard_timer);

		if ( ectp_timerlist_add(timer->list, timer) < 0 )
			err_goto("timer_dispatchnsmit() : ectp_timerlist_add() failed.",
					discard_timer);

		if ( ectp_skb_retransmit(skb) < 0 )
			err_msg("ectp_timer_dispatch() : ectp_skb_retransmit() failed.");
		break;

	case ECTP_TIMER_HGT:
		break;

	case ECTP_TIMER_CCT:
		break;

	case ECTP_TIMER_CWT:
		break;
	}

	return;

discard_timer:
	ectp_ti_free(timer);
	return;

}

