/* Enhanced Communication Transport Protocol Part-3 Implementation
 *
 * This file is part of the ECTP-3 over UDP Implementation.
 *
 * user.c
 * 	-	the implementation of the ECTP-3 user API.
 *
 * Please send any bug reports or fixes you make to the email address(es):
 * 	ECTP-3 developers <saintpaul1978@mail.knu.ac.kr>
 *
 * Written or modified by:
 * 	Sang Tae Kim				<saintpaul1978@mail.knu.ac.kr>
 *	Jae Sung Park				<knucsid@gmail.com>
 *
 */

#include <ectp/ectp.h>
#include <ectp/types.h>
#include <ectp/structs.h>
#include <ectp/user.h>
#include <ectp/event.h>

/**
 * Create ECTP-3 Socket
 * - creates ECTP-3 endpoint for communications and return descriptor.
 * - On success, a new socket descriptor for ECTP-3 is returned.
 *   On error, -1 is returned.
 */
int msocket( int domain, int role )
{
	struct ectp_socket *eso = NULL;

	/* Two process:
	 * 	1.	create a new socket and add it to the global socket list.
	 * 	2.	create a new protocol endpoint.
	 */

	eso = ectp_create_socket( domain, (ectp_role_t) role );
	if ( !eso ) {
		err_ret("msocket() : ectp_create_socket() failed.", -1);
	}

	if ( ectp_create_endpoint( eso, (ectp_role_t) role ) < 0 ) {
		ectp_destroy_socket( eso );
		err_ret("msocket() : ectp_create_endpoint() failed.", -1);
	}

	/*
	 * Socket created! (STATE: CLOSED)
	 */
	fputs("ECTP-3, Socket Created! (STATE: CLOSED)\n", stdout);

	return eso->esd;
}

/**
 * Connection Creation (TO)
 */
int mcreate( int esd, struct sockaddr_storage *local_addr,
				 struct sockaddr_storage *group_addr, int time )
{
	struct ectp_socket	*eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mcreate() : invalid socket descriptor.", -1);

	if ( ectp_create_connection(eso, local_addr, group_addr, time) < 0 )
		err_ret("mcreate() : ectp_create_connection() failed.", -1);

	if ( ectp_create_tree(eso) < 0 )
		err_ret("mcreate() : ectp_create_tree() failed.", -1);

	/* Invoke main loop and event handlers. */
	if ( ectp_ops_start_events(eso) < 0 )
		err_ret("mcreate() : ectp_ops_start_events() failed.", -1);

	/* Handshake for Connection Creation. */
	if ( ectp_ops_connection_creation(eso, time) < 0 )
		err_ret("mcreate() : Initial Handshake failed.", -1);

	/*
	 * Connection created! (STATE: ESTABLISHED)
	 */
	
	puts("ECTP-3, Connection Creation Completed! (STATE: ESTABLISHED)");

	return 0;
}

/** 
 * Late Join (TU)
 *
 * local_addr : local address of TO
 */
int mjoin( int esd, struct sockaddr_storage *local_addr, 
			   struct sockaddr_storage *group_addr, int time )
{
	struct ectp_socket	*eso = NULL;
	int result = -1;

	/* Get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mjoin() : invalid socket descriptor.", -1);

	/* Check the role of late joiner. */
	if ( eso->ep->role != ECTP_ROLE_LE && eso->ep->role != ECTP_ROLE_LO ) {
		err_ret("mjoin() : Late joiner must be either LE or LO.", -1);
	}

	if ( ectp_create_connection(eso, local_addr, group_addr, time) < 0 )
		err_ret("mjoin() : ectp_create_connection() failed.", -1);

	if ( ectp_create_tree(eso) < 0 )
		err_ret("mjoin() : ectp_create_tree() failed.", -1);

	/* Invoke main loop and event handlers. */
	if ( ectp_ops_start_events(eso) < 0 )
		err_ret("mjoin() : ectp_ops_start_events() failed.", -1);

	/* Handshake for Late Join. */
	result = ectp_ops_start_latejoin(eso, time);
	if ( result < 0 )
		err_ret("mjoin() : Late Join Handshake failed.", -1);
		/* ectp_ops_stop_events(eso), tree, connection destroy.  */
	else if ( result == 0 )
		err_ret("mjoin() : Late Join is Denied by TO.", -1);
		/* ectp_ops_stop_events(eso), tree, connection destroy.  */

	/*
	 * Late Join Completed! (STATE: ESTABLISHED)
	 */

	return 0;
}

/*
 * Connection Termination (TO)
 */
int mterminate( int esd, int f )
{
	struct ectp_socket *eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mterminate() : invalid socket descriptor.", -1);

	/* Check the role of terminate */
	if ( eso->ep->role != ECTP_ROLE_TO )
		err_ret("mterminate() : Only TO can terminate the connection.", -1);

	/* pending */

	/* send ECTP_PKT_CT */
	if ( ectp_ops_conn_terminate(eso, f) < 0 )
		err_ret("mterminate() : ectp_ops_conn_terminate() failed.", -1);

	/*
	 * 	destory handler of TO 
	 * 	mainloop, event_handler, join_handler, token_handler.
	 */
	if ( ectp_ops_end_events(eso) < 0 )
		err_ret("mterminate() : ectp_ops_end_events() failed.", -1);
	

	/* destory structure of TO (create in reverse order) */
	if ( ectp_destroy_tree(eso) < 0 )
		err_ret("mterminate() : ectp_destroy_tree() failed.", -1);
	
	if ( ectp_destroy_connection(eso) < 0 )
		err_ret("mterminate() : ectp_destroy_connection() failed.", -1);

	if ( ectp_destroy_endpoint(eso->ep) < 0 )
		err_ret("mterminate() : ectp_destroy_endpoint() failed.", -1);

	if ( ectp_destroy_socket(eso) < 0 )
		err_ret("mterminate() : ectp_destroy_socket() failed.", -1);

	printf( "TO Terminate\n" );
	return 0;
}


/**
 * User Leave (TO/LO/LE)
 *
 * If to is NULL, the local endpoint starts 'User Leave' process. If to is the
 * specified user address, parent can force the user to leave the connection.
 *
 * RETURN:
 * -	On success, 0 is returned.
 *  	Otherwise, -1 is returned.
 */
int mleave( int esd, const struct sockaddr_storage *to, socklen_t tolen )
{
	struct ectp_socket *eso = NULL;
	struct ectp_endpoint *ep = NULL;

	/* get current socket for the given 'esd' */
	if ( !(eso = ectp_get_socket(esd)) )
		err_ret("mleave() : invalid socket descriptor.", -1);

	ep = eso->ep;

	switch ( ep->role ) {

	case ECTP_ROLE_TO:	
		/* TO --> Troublemaker Ejection. */

		if ( ectp_ops_user_leave(eso, (ectp_addr_t *) to, tolen) < 0 )
			err_ret("mleave() : ectp_ops_user_leave() failed.", -1);
		break;

	case ECTP_ROLE_LO:
		/* LO --> User Leave Operation. */
		if ( !to ) {
			if ( ectp_ops_user_leave(eso, NULL, 0) < 0 )
				err_ret("mleave() : ectp_ops_user_leave() failed.", -1);

			/* If normal termination, it doesn't have to send LR to all
			 * 	 the children. If not normal, it may send LR to all the
			 * 	 children by need. Otherwise, children nodes should
			 * 	 consider HB/HB-ACK to detect parent's failure. */

			/* invoke shutdown */
		}

		/* LO --> Troublemaker Ejection. */
		else {
			if ( ectp_ops_user_leave(eso, (ectp_addr_t *) to, tolen) < 0 )
				err_ret("mleave() : ectp_ops_user_leave() failed.", -1);
		}
		break;

	case ECTP_ROLE_LE:
		/* LE --> User Leave Operation. */
		if ( ectp_ops_user_leave(eso, NULL, 0) < 0 )
			err_ret("mleave() : ectp_ops_user_leave() failed.", -1);
		
		/* invoke shutdown */

		
		/* destory handler - mainloop, event_handler */
		if ( ectp_ops_end_events(eso) < 0 )
			err_ret("mleave() : ectp_ops_end_events() failed.", -1);


		/* destory structure of LE (create in reverse order) */
		if ( ectp_destroy_connection(eso) < 0 )
			err_ret("mleave() : ectp_destroy_connection() failed.", -1);

		if ( ectp_destroy_endpoint(eso->ep) < 0 )
			err_ret("mleave() : ectp_destroy_endpoint() failed.", -1);

		if ( ectp_destroy_socket(eso) < 0 )
			err_ret("mleave() : ectp_destroy_socket() failed.", -1);

		break;

	default:
		err_ret("mleave() : invalid role.", -1);
	}

	return 0;
}

/**
 * Token Get (LE/LO)
 *
 * NOTE:
 * 		Blocking function!
 *
 */
int mtoken_get( int esd )
{
	struct ectp_socket	*eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mtoken_get() : invalid socket descriptor.", -1);

	/* Check the role. */ 
	if ( eso->ep->role != ECTP_ROLE_LE && eso->ep->role != ECTP_ROLE_LO ) {
		err_ret("mtoken_get() : Either LE or LO can request a token.", -1);
	}

	return ectp_ops_request_tokenget(eso);
}

/**
 * Token Return (LE/LO)
 *
 * NOTE:
 * 		Blocking function!
 *
 */
int mtoken_return( int esd )
{
	struct ectp_socket	*eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mtoken_return() : invalid socket descriptor.", -1);

	/* Check the role. */
	if ( eso->ep->role != ECTP_ROLE_LE && eso->ep->role != ECTP_ROLE_LO ) {
		err_ret("mtoken_return() : Either LE or LO can retuen a token.", -1);
	}

	return ectp_ops_request_tokenreturn(eso);
}

/**
 * TO use msendm() to transport a multicast data 'msg' to TUs.
 */
ssize_t msendm( int esd, const void *msg, size_t len )
{
	struct ectp_socket	*eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("msendm() : invalid socket descriptor.", -1);

	/* Check the role of multicast data sender. */
	if ( eso->ep->role != ECTP_ROLE_TO && eso->ep->role != ECTP_ROLE_LO ) {
		err_ret("msendm() : Multicast sender must be either TO or LO.", -1);
	}

	return ectp_ops_sendmsg( eso, msg, len, ECTP_MCAST,
			&eso->group_addr, ectp_addrlen(eso->domain) );
}

/** 
 * SU use msendmsg() to transport a unicast data 'msg' to TO or LO.
 */
ssize_t msendmsg( int esd, const void *msg, size_t len, int flags,
					  const struct sockaddr_storage *to, socklen_t tolen )
{
	struct ectp_socket	*eso = NULL;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("msendmsg() : invalid socket descriptor.", -1);

	/* Check the capability of the unicast data sender. */
	if ( !ectp_ep_is_su(eso->ep) )
		err_ret("msendmsg() : Unicast sender must be SU. "
				"Please request a token.", -1);

	return ectp_ops_sendmsg(eso, msg, len, ECTP_UCAST,
			(ectp_addr_t *) to, tolen);
}

/** 
 *
 */
ssize_t mrecvmsg( int esd, void *msg, size_t len, int flags,
					  struct sockaddr_storage *from, socklen_t *fromlen )
{
	struct ectp_socket	*eso = NULL;
	ssize_t	nbytes = 0;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mrecvmsg() : invalid socket descriptor.", -1);

	/* Ths upplayer application does not use flags field currently.
	 * 
	 * 	ECTP_MCAST
	 * 	ECTP_UCAST
	 * 	ECTP_DONTWAIT	- using ectp_sem_trywait()
	 */

	if ( flags != 0 )
		err_ret("mrecvmsg() : flags field must be 0.", -1);

	switch ( eso->ep->role ) {

	case ECTP_ROLE_TO:		/* TO can only receive unicast data. */
		nbytes = ectp_ops_recvmsg( eso, msg, len, ECTP_UCAST,
					(ectp_addr_t *) from, fromlen );
		if ( nbytes < 0 )
			err_ret("mrecvmsg() : ectp_ops_recvmsg() failed.", -1);
		break;

	case ECTP_ROLE_LO:		/* LO(RA) can receive both multicast */
		break;					/* and unicast data. */

	case ECTP_ROLE_LE:		/* LE(TU) can only receive multicast data. */
		nbytes = ectp_ops_recvmsg( eso, msg, len, ECTP_MCAST,
					(ectp_addr_t *) from, fromlen );
		if ( nbytes < 0 )
			err_ret("mrecvmsg() : ectp_ops_recvmsg() failed.", -1);
		break;

	default:
		err_ret("mrecvmsg() : invalid role.", -1);
		break;

	} /* switch ( eso->ep->role ) */

	return nbytes;	/* return received msg size in bytes. */
}


int msetsockopt( int esd, int level,
					 int optname, void *optval, socklen_t optionlen )
{
	struct ectp_socket	*eso = NULL;
	int state = 0;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if( !eso )
		err_ret( "msetsockopt() : invalid socket descriptor.", -1 );

	if( !eso->ep )
		err_ret( "msetsockopt() : invalid end point.", -1 );

	if( !eso->ep->cn )
		err_ret( "msetsockopt() : invalid connection.", -1 );

	if( level == IPPROTO_ECTP ) {	/* Option of ECTP-3 */

		switch( optname ) {

		case ECTP_OPT_TCO:
			eso->ep->cn->tco = *(__u8*)optval;
			break;

		case ECTP_OPT_AGN:
			eso->ep->cn->agn = *(__u8*)optval;
			break;

		case ECTP_OPT_MSS:
			eso->ep->cn->mss = *(__u16*)optval;
			break;

		case ECTP_OPT_EVENTS:
			ectp_update_callback(
				&eso->callback, (struct ectp_callback*) optval);
			break;

		default:
			state = -1;
			break;
		} /* switch( optname ) */
	} /* if( level == IPPROTO_ECTP ) */

	else {
		state = setsockopt( esd, level, optname, optval, optionlen );
	}

	if( state == -1 )
		return ECTP_FAILURE;

	return ECTP_SUCCESS;
}


int mgetsockopt( int esd, int level,
					 int optname, void *optval, socklen_t *optionlen)
{
	struct ectp_socket *eso = NULL;
	struct ectp_connection *cn = NULL;
	int state = 0;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	cn = eso->ep->cn;

	if ( !eso )
		err_ret( "mgetsockopt() : invalid socket descriptor.", -1 );

	if( !eso->ep )
		err_ret( "mgetsockopt() : invalid end point.", -1 );

	if ( !eso->ep->cn )
		err_ret( "mgetsockopt() : invalid connection.", -1 );


	if( level == IPPROTO_ECTP ) {

		switch( optname ) {


		case ECTP_OPT_TCO:
			*(__u8 *) optval = cn->tco;
			*optionlen = 1;	/* Original length : 2bit */
			break;

		case ECTP_OPT_AGN:
			*(__u8 *) optval = cn->agn;
			*optionlen = sizeof(cn->agn);
			break;

		case ECTP_OPT_MSS:
			*(__u16 *) optval = cn->mss;
			*optionlen = sizeof(cn->mss);
			break;

		default:
			state = -1;
			break;
		} /* if( level == IPPROTO_ECTP ) */
	} /* switch( optname ) */

	else {
		state = getsockopt( eso->m_sd, level, optname, optval, optionlen );
		state = getsockopt( eso->u_sd, level, optname, optval, optionlen );
	}

	if( state == -1 )
		return ECTP_FAILURE;

	return ECTP_SUCCESS;
}


int mjoin_confirm( int esd, int accept,
					   const struct sockaddr_storage *to, socklen_t tolen )
{
	struct ectp_socket	*eso = NULL;
	int result = -1;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mjoin_confirm() : invalid socket descriptor.", -1);


	if ( !accept )
		result = ectp_ops_deny_latejoin(eso, (ectp_addr_t *) to);
	else
		result = ectp_ops_accept_latejoin(eso, (ectp_addr_t *) to);


	if ( result < 0 )
		err_ret("mjoin_confirm() : error on making JC.", -1);

	return 0;
}


int mtoken_confirm( int esd, int accept,
						const struct sockaddr_storage *to, socklen_t tolen )
{
	struct ectp_socket	*eso = NULL;
	int result;

	/* get current socket for the given 'esd' */
	eso = ectp_get_socket(esd);
	if ( !eso )
		err_ret("mtoken_confirm() : invalid socket descriptor.", -1);


	if ( !accept )	/* deny */
		result = ectp_ops_deny_tokenget(eso, (ectp_addr_t *) to);
	else 
		result = ectp_ops_accept_tokenget(eso, (ectp_addr_t *) to);


	if ( result < 0 )
		err_ret("mtoken_confirm() : error on making TGC.", -1);

	return 0;
}

/* vi: set nonu:ts=4:sw=4: */
